import random
import itertools
import numpy
from lclpy.localsearch.move.abstract_move \
    import AbstractMove


class MultiNeighbourhood(AbstractMove):
    """Implements a class that can be used to combine different move functions.

    Moves are represented in the following way: \n
    (index move function, a representation of a move from said move function)\n
    Note that all move functions must work on the same type of data structure.
    Combining move functions meant for different data structures won't work.

    Parameters
    ----------
    move_func_list : list of AbstractMove or tuple of AbstractMove
        A list or tuple that contains the move functions one wishes to combine.
    weights : list of int or tuple of int, optional
        A list or tuple that contains weights to increase or decrease the
        chance of a move function being used to generate a random move. The
        weight for a function must have the same index as the function.
        Weights are positive numbers and need to be defined for every move
        function. In the default case, every function will have the same
        weight. Note that it's perfectly possible to pass the probabilities
        instead of the weights.

    Attributes
    ----------
    _move_func_list : tuple of AbstractMove
        Contains all used move functions.
    _size : int
        The size of _move_func_list.
    _tresholds : tuple of float
        Tresholds associated with the move functions.

    Examples
    --------
    Get all possible moves, you should NEVER do this. You should evaluate only
    one move at a time. This example is simply to show the behaviour of
    get_moves and how to perform and undo a move:

    .. doctest::

        >>> import numpy
        >>> from lclpy.localsearch.move.array_swap import ArraySwap
        >>> from lclpy.localsearch.move.array_reverse_order \\
        ...     import ArrayReverseOrder
        >>> from lclpy.localsearch.move.multi_neighbourhood \\
        ...     import MultiNeighbourhood
        ... # init array, a move will be performed on this array
        >>> array = numpy.array([0, 1, 2])
        ... # init
        >>> swap = ArraySwap(len(array))
        >>> reverse = ArrayReverseOrder(len(array))
        >>> move_func_list = [swap, reverse]
        >>> multi = MultiNeighbourhood(move_func_list)
        ... # get all possible moves in all_moves
        >>> all_moves = []
        >>> for move in multi.get_moves():
        ...     all_moves.append(move)
        >>> all_moves
        [(0, (0, 1)), (0, (0, 2)), (0, (1, 2)), (1, (0, 1)), (1, (0, 2)), (1, (1, 2))]
        >>> # picking an arbitrary move
        >>> # Never pick a move like this yourself. It only is done here for
        >>> # the sake of showing you a clear example.
        >>> a_move = all_moves[2]
        >>> a_move
        (0, (1, 2))
        >>> # performing the move on the array
        >>> multi.move(array, a_move)
        >>> array
        array([0, 2, 1])
        >>> # undoing the move on the array
        >>> multi.undo_move(array, a_move)
        >>> array
        array([0, 1, 2])

    An example of generating some random moves with get_random_move with
    default weights (every move function has the same chance of being chosen):

    .. doctest::

        >>> import random
        >>> from lclpy.localsearch.move.array_swap import ArraySwap
        >>> from lclpy.localsearch.move.array_reverse_order \\
        ...     import ArrayReverseOrder
        >>> from lclpy.localsearch.move.multi_neighbourhood \\
        ...     import MultiNeighbourhood
        ... # set seed random
        ... # not needed, is only used here to always get the same moves.
        >>> random.seed(0)
        ... # init
        >>> swap = ArraySwap(10)
        >>> reverse = ArrayReverseOrder(10)
        >>> move_func_list = [swap, reverse]
        >>> multi = MultiNeighbourhood(move_func_list)
        ... # tests
        >>> multi.get_random_move()
        (1, (0, 6))
        >>> multi.get_random_move()
        (0, (7, 8))
        >>> multi.get_random_move()
        (0, (4, 7))
        >>> multi.get_random_move()
        (0, (3, 8))

    An example of generating some random moves with get_random_move with
    defined weights (swap weight: 1, reverse weight: 3):

    .. doctest::

        >>> import random
        >>> from lclpy.localsearch.move.array_swap import ArraySwap
        >>> from lclpy.localsearch.move.array_reverse_order \\
        ...     import ArrayReverseOrder
        >>> from lclpy.localsearch.move.multi_neighbourhood \\
        ...     import MultiNeighbourhood
        ... # set seed random
        ... # not needed, is only used here to always get the same moves.
        >>> random.seed(0)
        ... # init
        >>> swap = ArraySwap(10)
        >>> reverse = ArrayReverseOrder(10)
        >>> move_func_list = [swap, reverse]
        >>> weights = [1, 3]
        >>> multi = MultiNeighbourhood(move_func_list, weights)
        ... # tests
        >>> multi.get_random_move()
        (1, (0, 6))
        >>> multi.get_random_move()
        (1, (7, 8))
        >>> multi.get_random_move()
        (1, (4, 7))
        >>> multi.get_random_move()
        (1, (3, 8))
        >>> multi.get_random_move()
        (0, (1, 2))

    An example of generating some random moves with get_random_move with
    defined probabilities (swap probability: 0.25, reverse probability: 0.75).
    Note that the results are exactly the same as the results of the previous
    doctest:

    .. doctest::

        >>> import random
        >>> from lclpy.localsearch.move.array_swap import ArraySwap
        >>> from lclpy.localsearch.move.array_reverse_order \\
        ...     import ArrayReverseOrder
        >>> from lclpy.localsearch.move.multi_neighbourhood \\
        ...     import MultiNeighbourhood
        ... # set seed random
        ... # not needed, is only used here to always get the same moves.
        >>> random.seed(0)
        ... # init
        >>> swap = ArraySwap(10)
        >>> reverse = ArrayReverseOrder(10)
        >>> move_func_list = [swap, reverse]
        >>> weights = [0.25, 0.75]
        >>> multi = MultiNeighbourhood(move_func_list, weights)
        ... # tests
        >>> multi.get_random_move()
        (1, (0, 6))
        >>> multi.get_random_move()
        (1, (7, 8))
        >>> multi.get_random_move()
        (1, (4, 7))
        >>> multi.get_random_move()
        (1, (3, 8))
        >>> multi.get_random_move()
        (0, (1, 2))

    Test of using delta evaluation with a multineighbourhood:

    .. doctest::

        >>> import numpy
        >>> from lclpy.localsearch.move.array_swap import ArraySwap
        >>> from lclpy.localsearch.move.array_reverse_order \\
        ...     import ArrayReverseOrder
        >>> from lclpy.localsearch.move.multi_neighbourhood \\
        ...     import MultiNeighbourhood
        >>> from lclpy.evaluation.quadratic_assignment_evaluation_function \\
        ...     import QuadraticAssignmentEvaluationFunction
        ... # init array, a move will be performed on this array
        >>> array = numpy.array([0, 1, 2, 3])
        ... # init multineighbourhood
        >>> swap = ArraySwap(len(array))
        >>> reverse = ArrayReverseOrder(len(array))
        >>> move_func_list = [swap, reverse]
        >>> multi = MultiNeighbourhood(move_func_list)
        ... # init evaluation function
        >>> dist_matrix = numpy.array(
        ... [[0, 2, 9, 5],
        ...  [2, 0, 4, 6],
        ...  [9, 4, 0, 3],
        ...  [5, 6, 3, 0]])
        ... # init flow matrix
        >>> flow_matrix = numpy.array(
        ... [[0, 2, 0, 0],
        ...  [2, 0, 4, 0],
        ...  [0, 4, 0, 8],
        ...  [0, 0, 8, 0]])
        ... # init evaluation function
        >>> eval_func = QuadraticAssignmentEvaluationFunction(dist_matrix,
        ...                                                   flow_matrix,
        ...                                                   multi)
        ... # tests
        >>> order = [0, 1, 2, 3]
        >>> eval_func.evaluate([0, 1, 2, 3])
        44
        >>> eval_func.evaluate([0, 1, 3, 2])
        52
        >>> eval_func.delta_evaluate(order, (0, (2, 3)))
        8
        >>> eval_func.delta_evaluate(order, (1, (2, 3)))
        8
        >>> eval_func.evaluate([0, 3, 2, 1])
        54
        >>> eval_func.delta_evaluate(order, (0, (1, 3)))
        10
        >>> eval_func.delta_evaluate(order, (1, (1, 3)))
        10
        >>> eval_func.evaluate([3, 2, 1, 0])
        38
        >>> eval_func.delta_evaluate(order, (1, (0, 3)))
        -6

    """

    def __init__(self, move_func_list, weights=None):
        super().__init__()

        # init
        self._move_func_list = tuple(move_func_list)

        self._size = len(move_func_list)

        if weights is None:
            weights = numpy.ones(self._size)

        # calculate weight

        total_weight = sum(weights)

        # calculate probabilities

        probabilities = numpy.empty(self._size)

        for i in range(self._size):

            probabilities[i] = weights[i] / total_weight

        self._tresholds = numpy.array(
            list(itertools.accumulate(probabilities)))

    def get_move_type(self):
        """Returns the move type.

        Returns
        -------
        str
            The move type.

        """

        return 'multi_neighbourhood'

    def move(self, data, move):
        """Performs a move.

        Parameters
        ----------
        data
            The dataset that is being explored. It will be altered after the
            method call.
        move : tuple of int
            A representation of a valid move.

        """

        # pick the right move_function and perform the move
        self._move_func_list[move[0]].move(data, move[1])

    def undo_move(self, data, move):
        """Undoes a move.

        Make sure you only undo a move after it's has been performed.

        Parameters
        ----------
        data
            The dataset that is being explored. It will be altered after the
            method call.
        move : tuple of int
            A representation of the move one wishes to undo.

        """

        # pick the right move_function and undo the move
        self._move_func_list[move[0]].undo_move(data, move[1])

    def get_moves(self):
        """A generator used to return all valid moves in the neighbourhood.

        Note that all moves of all neighbourhoods are included. Duplicate moves
        are possible. All moves are generated for one neighbourhood, after
        which they will all be generated for next neighbourhood until no
        remain. The neighbourhoods will be visited in the order their
        move functions had in move_func_list when it was passed to the
        constructor.

        Yields
        ------
        tuple of int
            The next valid move.

        """

        # iterating over all move functions
        for i in range(self._size):

            # get the neighbourhood from the move_function
            neighbourhood = self._move_func_list[i].get_moves()

            # yield all moves from said neighbourhood
            for move in neighbourhood:
                yield (i, move)

    def get_random_move(self):
        """A method used to generate a random move in the neighbourhood.

        Returns
        -------
        tuple of int
            A random valid move.

        """

        # generate a random number
        random_number = random.random()

        # compare random number with all tresholds, expept the last.
        for i in range(self._size - 1):

            # if the random number is smaller than the treshold,
            # the move function that corresponds to the threshold is chosen to
            # generate the random move

            if(random_number < self._tresholds[i]):
                return (i, self._move_func_list[i].get_random_move())

        # The only treshold that we haven't checked yet is the last one, but
        # this one will always be one, thus it will always be bigger than our
        # random number. Therefore we can return a move from the corresponding
        # move function without checks.

        return (self._size - 1,
                self._move_func_list[self._size - 1].get_random_move())

    def size(self):
        """Function to get amount of neighbourhoods in the multi neighbourhood.

        Returns
        -------
        int
            The amount of neighbourhoods.

        """

        return len(self._move_func_list)

    def select_get_moves(self, neighbourhood_nr):
        """A generator used to return all moves from a specific neighbourhood.

        Yields
        ------
        tuple of int
            The next valid move from the specified neighbourhood.

        """

        # get the neighbourhood from the _move_func_list
        neighbourhood = self._move_func_list[neighbourhood_nr].get_moves()

        # yield all moves from said neighbourhood
        for move in neighbourhood:
            yield (neighbourhood_nr, move)

    def select_random_move(self, neighbourhood_nr):
        """A method used to generate a random move from a specific neighbourhood.

        Parameters
        ----------
        neighbourhood_nr : int
            Number of the neighbourhood. This number is the index of the
            neighbourhood in the list of move functions given to the
            constructor.

        Returns
        -------
        tuple of int
            A random valid move from the specified neighbourhood.

        """

        return (neighbourhood_nr,
                self._move_func_list[neighbourhood_nr].get_random_move())



"""This package contains the classes and packages related to localsearch.

- acceptance : package for acceptance functions
- move : package for move functions

Local search algorithms:

- simulatedannealing : package for everything unique to simulated annealing
- steepestdescent : package for everything unique to steepest descent
- tabusearch : package for everything unique to tabu search
- vns : package for everything unique to variable neighbourhood search (VNS)

Note that all the local search algorithms use delta-evaluation.
If your evaluation function does not have it implemented, errors will be
raised.

When using moves and evaluation functions implemented by this library delta
evaluation can be easily enabled, by passing the move function as a parameter
to the evaluation function:

.. code-block:: python

    move = ImplementedMove(size)
    eval_func = \\
    ImplementedEvaluationFunction(parameter_1, parameter_2, move_function=move)

Note that the evaluation function might have less or more parameters than in
this example.

You can find information how to implement your own delta evaluation in the
docstring of lclpy.evaluation.deltaeval.__init__ or in the documentation of
lclpy.evaluation.deltaeval.

If you don't feel like implementing delta_evaluation for your own evaluation
and or move classes, for whatever reason, you can always do the implement
evaluate_move like this in your Problem class:

.. code-block:: python

    def evaluate_move(self, move):

        self.move(move)
        new_value = self._evaluation_function.evaluate(self._order)
        self.undo_move(move)

        return new_value - self._evaluation_function.evaluate(self._order)

If you want to use this method with an existing problem class from this
library, it's the best to define your own subclass that overwrites the default
behaviour of the method and use this subclass:

.. code-block:: python

    class MyProblemClass(ImplementedProblemClass):
        def evaluate_move(self, move):

            self.move(move)
            new_value = self._evaluation_function.evaluate(self._order)
            self.undo_move(move)

            return new_value - self._evaluation_function.evaluate(self._order)

Do remember not to pass a move function to the evaluation function's
constructor if you use these approaches.
Also note that there will be a significant slowdown.

"""



from abc import ABC, abstractmethod


class AbstractAcceptanceFunction(ABC):
    """Template to create acceptance functions.

    Acceptance functions determine if a worse state than the current
    state is accepted as the new current state.

    """

    def __init__(self):
        super().__init__()

    @abstractmethod
    def accept(self, delta_value):
        """Checks if the worse state will be accepted

        Parameters
        ----------
        delta_value : int or float
            The difference in quality between the current state and the
            potential new state. This can be calculated by substracting the
            quality of the current state from the quality of the potential new
            state.

        Returns
        -------
        bool
            The function returns true if the problem is accepted,
            the function returns false if the problem is rejected.

        """
        pass



from lclpy.problem.abstract_local_search_problem \
    import AbstractLocalSearchProblem
import numpy
from lclpy.aidfunc.error_func import not_multi_move_type
from lclpy.aidfunc.error_func import NoNextNeighbourhood

from statistics import mean, StatisticsError


class ArrayProblem(AbstractLocalSearchProblem):
    """Contains all the data needed to handle a problem.

    Parameters
    ----------
    evaluation_function : AbstractEvaluationFunction
        The evaluation function that needs to be used for the problem.
    move_function : AbstractMove
        The move function that needs to be used for the problem.
    order : numpy.ndarray or list, optional
        A one dimensional array that contains the order of the points to start
        with. All values are int, unique and are within the interval
        [0,size[.
        The default value is None.
        In the default case a numpy array will be generated. The generated
        array's values will always be ordered from small to big.

    Attributes
    ----------
    _evaluation_function : AbstractEvaluationFunction.
        The evaluation function that is used for the problem
    _move_function : AbstractMove
        The move function that is used for the problem.
    _order : numpy.ndarray
        A 1 dimensional array that contains the current order of the points.
        All values are int, unique and are within the interval [0,size[.
    _starting_order : numpy.ndarray
        The initial value of _order.
    best_order : numpy.ndarray
        Contains the order of the best found problem.
    best_order_value: int or float
        The evaluation value of the best found problem.

    Examples
    --------

    A simple example, demonstrates the use of move, undo_move, evaluate and
    set_as_best. Note that problem.set_as_best_order does NOT check if the
    value actually belongs to the order NOR does it check if the value is
    better than the previous best value:

    .. doctest::

        >>> import numpy
        >>> from lclpy.localsearch.move.tsp_array_swap import TspArraySwap
        >>> from lclpy.evaluation.tsp_evaluation_function \\
        ...     import TspEvaluationFunction
        >>> from lclpy.problem.array_problem import ArrayProblem
        ... # init distance matrix
        >>> distance_matrix = numpy.array(
        ... [[0, 2, 5, 8],
        ...  [2, 0, 4, 1],
        ...  [5, 4, 0, 7],
        ...  [8, 1, 7, 0]])
        ... # init move function
        >>> size = distance_matrix.shape[0]
        >>> move_func = TspArraySwap(size)
        ... # init evaluation function
        >>> evaluation_func = TspEvaluationFunction(distance_matrix)
        ... # init problem
        >>> problem = ArrayProblem(evaluation_func, move_func, size)
        ... # default generated order
        >>> problem._order
        array([0, 1, 2, 3])
        >>> # evaluating the current order
        >>> value = problem.evaluate()
        >>> value
        21
        >>> # saving the current order as the best order.
        >>> problem.set_as_best(value)
        ... # get the best order and it's value
        >>> problem.best_order
        array([0, 1, 2, 3])
        >>> problem.best_order_value
        21
        >>> # perform a move and evaluate the new order
        >>> problem.move((2,3))
        >>> problem._order
        array([0, 1, 3, 2])
        >>> value = problem.evaluate()
        >>> value
        15
        >>> # saving the current order as the best order
        >>> problem.set_as_best(value)
        >>> problem.best_order
        array([0, 1, 3, 2])
        >>> problem.best_order_value
        15
        >>> # undoing move and rechecking value.
        >>> # Note that best_order and best_order_value don't change.
        >>> problem.undo_move((2,3))
        >>> problem._order
        array([0, 1, 2, 3])
        >>> problem.best_order
        array([0, 1, 3, 2])
        >>> problem.best_order_value
        15

    Initialising with a non-default order:

    .. doctest::

        >>> import numpy
        >>> from lclpy.localsearch.move.tsp_array_swap import TspArraySwap
        >>> from lclpy.evaluation.tsp_evaluation_function \\
        ...     import TspEvaluationFunction
        >>> from lclpy.problem.array_problem import ArrayProblem
        ... # init distance matrix
        >>> distance_matrix = numpy.array(
        ... [[0, 2, 5, 8],
        ...  [2, 0, 4, 1],
        ...  [5, 4, 0, 7],
        ...  [8, 1, 7, 0]])
        ... # wanted order is the order we want.
        >>> wanted_order = numpy.array([0, 3, 2, 1])
        ... # init move function
        >>> size = distance_matrix.shape[0]
        >>> move_func = TspArraySwap(size)
        ... # init evaluation function
        >>> evaluation_func = TspEvaluationFunction(distance_matrix)
        ... # init problem
        >>> problem = ArrayProblem(
        ...     evaluation_func, move_func, size, wanted_order)
        ... # the order of the problem
        >>> problem._order
        array([0, 3, 2, 1])

    """

    def __init__(self, evaluation_function, move_function, size, order=None):
        super().__init__()

        # init variables
        self._evaluation_function = evaluation_function
        self._move_function = move_function

        if self._move_function.get_move_type() is not 'multi_neighbourhood':
            self.first_neighbourhood = not_multi_move_type
            self.next_neighbourhood = not_multi_move_type
            self.previous_neighbourhood = not_multi_move_type
            self.select_get_moves = not_multi_move_type
            self.select_random_move = not_multi_move_type
        else:
            self.current_neighbourhood = 0
            self.neighbourhood_size = move_function.size()

        if order is None:
            self._order = numpy.arange(size)
        else:
            self._order = numpy.array(order)

        self._starting_order = numpy.array(self._order)

    def move(self, move):
        """Performs a move on _order.

        Parameters
        ----------
        move_number : tuple of int
            Represents a unique valid move.

        """

        self._move_function.move(self._order, move)

    def undo_move(self, move):
        """Undoes a move on _order .

        Parameters
        ----------
        move_number : tuple of int
            Represents a unique valid move.

        """

        self._move_function.undo_move(self._order, move)

    def get_moves(self):
        """An iterable that returns all valid moves in the complete neighbourhood.

        Yields
        -------
        tuple of int
            The next move in the neighbourhood.

        Examples
        --------
        Gets all moves from the neigbourhood, you should NEVER do this. You
        should evaluate only one move at a time. This example is simply to
        show the behaviour of get_moves.

        .. doctest::

            >>> import numpy
            >>> from lclpy.localsearch.move.tsp_array_swap \\
            ...     import TspArraySwap
            >>> from lclpy.evaluation.tsp_evaluation_function \\
            ...     import TspEvaluationFunction
            >>> from lclpy.problem.array_problem import ArrayProblem
            ... # init distance matrix
            >>> distance_matrix = numpy.array(
            ... [[0, 2, 5, 8],
            ...  [2, 0, 4, 1],
            ...  [5, 4, 0, 7],
            ...  [8, 1, 7, 0]])
            ... # init move function
            >>> size = distance_matrix.shape[0]
            >>> move_func = TspArraySwap(size)
            ... # init evaluation function
            >>> evaluation_func = TspEvaluationFunction(distance_matrix)
            ... # init problem
            >>> problem = ArrayProblem(evaluation_func, move_func, size)
            ... # retrieve all moves with get_moves
            >>> all_moves = []
            >>> for move in problem.get_moves():
            ...     all_moves.append(move)
            >>> all_moves
            [(1, 2), (1, 3), (2, 3)]

        """

        return self._move_function.get_moves()

    def get_random_move(self):
        """A function to return a random move from the complete neighbourhood.

        Returns
        -------
        tuple of int
            Represents one unique valid move in the neighbourhood.

        Examples
        --------
        Get a random valid move:

        .. doctest::

            >>> import numpy
            >>> from lclpy.localsearch.move.tsp_array_swap \\
            ...     import TspArraySwap
            >>> from lclpy.evaluation.tsp_evaluation_function \\
            ...     import TspEvaluationFunction
            >>> from lclpy.problem.array_problem import ArrayProblem
            ... # init distance matrix
            >>> distance_matrix = numpy.array(
            ... [[0, 2, 5, 8],
            ...  [2, 0, 4, 1],
            ...  [5, 4, 0, 7],
            ...  [8, 1, 7, 0]])
            ... # init move function
            >>> size = distance_matrix.shape[0]
            >>> move_func = TspArraySwap(size)
            ... # init evaluation function
            >>> evaluation_func = TspEvaluationFunction(distance_matrix)
            ... # init problem
            >>> problem = ArrayProblem(evaluation_func, move_func, size)
            ... # get a random move and check if it's in the neighboorhood.
            >>> move = problem.get_random_move()
            >>> move in [(1, 2), (1, 3), (2, 3)]
            True

        """

        return self._move_function.get_random_move()

    def evaluate_move(self, move):
        """Evaluates the quality gained or lost by a potential move.

        Can lead to considerable speedups. Is equivalent to a delta evaluation
        between _order and _order after the move is performed. Note that
        delta-evaluation needs to be implemented for the evaluation function
        and the move type for this method to work.

        Parameters
        ----------
        move : tuple of int
            Represents a unique valid move.

        Returns
        -------
        int or float
            The change in value of the eval-function if the move is performed.

        Examples
        --------
        A simple example:

        .. doctest::

            >>> import numpy
            >>> from lclpy.localsearch.move.tsp_array_swap \\
            ...     import TspArraySwap
            >>> from lclpy.evaluation.tsp_evaluation_function \\
            ...     import TspEvaluationFunction
            >>> from lclpy.problem.array_problem import ArrayProblem
            ... # init distance matrix
            >>> distance_matrix = numpy.array(
            ... [[0, 2, 5, 8],
            ...  [2, 0, 4, 1],
            ...  [5, 4, 0, 7],
            ...  [8, 1, 7, 0]])
            ... # init move function
            >>> size = distance_matrix.shape[0]
            >>> move_func = TspArraySwap(size)
            ... # init evaluation function
            >>> evaluation_func = TspEvaluationFunction(distance_matrix,
            ...                                         move_func)
            ... # init problem
            >>> problem = ArrayProblem(evaluation_func, move_func, size)
            ... # tests
            >>> problem.evaluate_move((1, 2))
            -3
            >>> problem.evaluate_move((1, 3))
            0
            >>> problem.evaluate_move((2, 3))
            -6

        """

        return self._evaluation_function.delta_evaluate(self._order, move)

    def evaluate(self):
        """A function to evaluate the current _order.

        Returns
        -------
        int or float
            An evaluation of the current state of _order.

        """

        return self._evaluation_function.evaluate(self._order)

    def set_as_best(self, evaluation_value):
        """Sets the current _order as the new best_order

        Parameters
        ----------
        evaluation_value : int or float
            The evaluation value of the current order. If you haven't kept or
            calculated said value, it can always be calculated with
            evaluate(). The recalculation will take time, however.

        """

        self.best_order = numpy.copy(self._order)
        self.best_order_value = evaluation_value

    def state(self):
        """Returns an immutable hashable object that describes the current state.

        Returns
        -------
        tuple
            A hashable object associated with the current state.

        Examples
        --------
        A simple example:

        .. doctest::

            >>> import numpy
            >>> from lclpy.localsearch.move.tsp_array_swap \\
            ...     import TspArraySwap
            >>> from lclpy.evaluation.tsp_evaluation_function \\
            ...     import TspEvaluationFunction
            >>> from lclpy.problem.array_problem import ArrayProblem
            ... # init distance matrix
            >>> distance_matrix = numpy.array(
            ... [[0, 2, 5, 8],
            ...  [2, 0, 4, 1],
            ...  [5, 4, 0, 7],
            ...  [8, 1, 7, 0]])
            ... # init move function
            >>> size = distance_matrix.shape[0]
            >>> move_func = TspArraySwap(size)
            ... # init evaluation function
            >>> evaluation_func = TspEvaluationFunction(distance_matrix,
            ...                                         move_func)
            ... # init problem
            >>> problem = ArrayProblem(evaluation_func, move_func, size)
            >>> problem.state()
            (0, 1, 2, 3)
            >>> problem.move((1, 3))
            >>> problem.state()
            (0, 3, 2, 1)

        """

        return tuple(self._order)

    def first_neighbourhood(self):
        """Changes the current neighbourhood to the first neighbourhood.

        Note that this function will only be useable if the neighbourhood given
        to the constructor is a MultiNeighbourhood.

        Raises
        ------
        WrongMoveTypeError
            If the move_function isn't a MultiNeighbourhood.

        """

        self.current_neighbourhood = 0

    def next_neighbourhood(self):
        """Changes the current neighbourhood to the next neighbourhood.

        Note that this function will only be useable if the neighbourhood given
        to the constructor is a MultiNeighbourhood.
        If this function is called when the last neighbourhood is the current
        neighbourhood, the last neighbourhood will remain the current
        neighbourhood and an exception will be raised.

        Raises
        ------
        NoNextNeighbourhood
            If there is no next neighbourhood. This is simply an indication
            that the current neighbourhood was the last neighbourhood.
        WrongMoveTypeError
            If the move_function isn't a MultiNeighbourhood.

        """

        self.current_neighbourhood += 1

        if self.current_neighbourhood is self.neighbourhood_size:
            self.current_neighbourhood -= 1
            raise NoNextNeighbourhood('There is no next neighbourhood.')

    def previous_neighbourhood(self):
        """Changes the current neighbourhood to the previous neighbourhood.

        Note that this function will only be useable if the neighbourhood given
        to the constructor is a MultiNeighbourhood.
        If this function is called when the first neighbourhood is the current
        neighbourhood, the first neighbourhood will remain the current
        neighbourhood.

        Raises
        ------
        WrongMoveTypeError
            If the move_function isn't a MultiNeighbourhood.

        """

        if self.current_neighbourhood is not 0:
            self.current_neighbourhood -= 1

    def select_get_moves(self):
        """Function to get all moves from the current neighbourhood.

        Note that this function will only be useable if the neighbourhood given
        to the constructor is a MultiNeighbourhood.

        Returns
        -------
        generator
            An iterable generator object that contains all the moves of the
            current neighbourhood.

        Raises
        ------
        WrongMoveTypeError
            If the move_function isn't a MultiNeighbourhood.

        """

        return self._move_function.select_get_moves(self.current_neighbourhood)

    def select_random_move(self):
        """A method used to generate a random move from the current neighbourhood.

        Note that this function will only be useable if the neighbourhood given
        to the constructor is a MultiNeighbourhood.

        Returns
        -------
        tuple of int
            A random valid move from the current neighbourhood.

        Raises
        ------
        WrongMoveTypeError
            If the move_function isn't a MultiNeighbourhood.

        """

        return self._move_function.select_random_move(
            self.current_neighbourhood)

    def reset(self):
        """Resets the object back to it's state after init.

        _order is replaced by a copy of _starting_order.

        Examples
        --------
        A simple example with the default order:

        .. doctest::

            >>> import numpy
            >>> from lclpy.localsearch.move.tsp_array_swap \\
            ...     import TspArraySwap
            >>> from lclpy.evaluation.tsp_evaluation_function \\
            ...     import TspEvaluationFunction
            >>> from lclpy.problem.array_problem import ArrayProblem
            ... # init distance matrix
            >>> distance_matrix = numpy.array(
            ... [[0, 2, 5, 8],
            ...  [2, 0, 4, 1],
            ...  [5, 4, 0, 7],
            ...  [8, 1, 7, 0]])
            ... # init move function
            >>> size = distance_matrix.shape[0]
            >>> move_func = TspArraySwap(size)
            ... # init evaluation function
            >>> evaluation_func = TspEvaluationFunction(distance_matrix,
            ...                                         move_func)
            ... # init problem
            >>> problem = ArrayProblem(evaluation_func, move_func, size)
            ... # tests
            >>> problem.state()
            (0, 1, 2, 3)
            >>> problem.move((1, 3))
            >>> problem.state()
            (0, 3, 2, 1)
            >>> problem.reset()
            >>> problem.state()
            (0, 1, 2, 3)
            >>> problem.move((1, 3))
            >>> problem.state()
            (0, 3, 2, 1)
            >>> problem.reset()
            >>> problem.state()
            (0, 1, 2, 3)

        An example with a non-default order:

        .. doctest::

            >>> import numpy
            >>> from lclpy.localsearch.move.tsp_array_swap \\
            ...     import TspArraySwap
            >>> from lclpy.evaluation.tsp_evaluation_function \\
            ...     import TspEvaluationFunction
            >>> from lclpy.problem.array_problem import ArrayProblem
            ... # init distance matrix
            >>> distance_matrix = numpy.array(
            ... [[0, 2, 5, 8],
            ...  [2, 0, 4, 1],
            ...  [5, 4, 0, 7],
            ...  [8, 1, 7, 0]])
            ... # init move function
            >>> size = distance_matrix.shape[0]
            >>> move_func = TspArraySwap(size)
            ... # init evaluation function
            >>> evaluation_func = TspEvaluationFunction(distance_matrix,
            ...                                         move_func)
            ... # init problem
            >>> problem = ArrayProblem(evaluation_func, move_func,
            ...                          size, [0, 1, 3, 2])
            ... # tests
            >>> problem.state()
            (0, 1, 3, 2)
            >>> problem.move((1, 3))
            >>> problem.state()
            (0, 2, 3, 1)
            >>> problem.reset()
            >>> problem.state()
            (0, 1, 3, 2)
            >>> problem.move((1, 3))
            >>> problem.state()
            (0, 2, 3, 1)
            >>> problem.reset()
            >>> problem.state()
            (0, 1, 3, 2)

        """

        self._order = numpy.array(self._starting_order)



"""This module contains all package specific Exceptions of lclpy.

The module also contains functions to raise Exceptions.

"""


def _not_implemented(*args):
    """An error raising function that can take any amount of parameters.

    Parameters
    ----------
    *args
        Variable length argument list, it does not matter what or how many
        parameters you use. Does not accept keyword arguments.

    Raises
    ------
    NotImplementedError
        This error will always be raised.

    """

    raise NotImplementedError


class WrongMoveTypeError(Exception):
    """Is raised when the wrong move type is used with a certain class."""
    pass


def not_multi_move_type(*args):
    """An error raising function that can take any amount of parameters.

    This function is meant to be called when someone uses a different move type
    than MultiNeighbourhood when one needs to use an instance of the
    MultiNeighbourhood class.

    Parameters
    ----------
    *args
        Variable length argument list, it does not matter what or how many
        parameters you use. Does not accept keyword arguments.

    Raises
    ------
    WrongMoveType
        This error will always be raised.

    """

    raise WrongMoveTypeError('This method requires you to use an object of the'
                             ' MultiNeighbourhood class as the move_function.')


class NoNextNeighbourhood(Exception):
    """Is raised when there is no next neighbourhood."""
    pass



from lclpy.aidfunc.error_func import _not_implemented


# base class for the tsp delta evaluation

class TSPDeltaEvaluate():
    """Class to perform delta-evaluation for TSP problems.

    Parameters
    ----------
    eval_func : AbstractEvaluationFunction
        The used evaluation function
    changed_distances
        This function returns the pairs who would have an altered evaluation
        value due to the move.
    next_to_current
        This function transforms the indices so that they can be used as
        indices in the unaltered array, yet return the value they would have
        had if the move was actually performed and they were used as indices.

    Attributes
    ----------
    eval_func : AbstractEvaluationFunction
        The used evaluation function
    changed_distances
        This function returns the pairs who would have an altered evaluation
        value due to the move.
    next_to_current
        This function transforms the indices so that they can be used as
        indices in the unaltered array, yet return the value they would have
        had if the move was actually performed and they were used as indices.

    Returns
    -------
    int or float
        The difference in quality if the move would be performed.

    """

    def __init__(self, eval_func, changed_distances, next_to_current):
        self.eval_func = eval_func
        self.changed_distances = changed_distances
        self.next_to_current = next_to_current

    def delta_evaluate(self, current_order, move):
        """Calculates the difference in quality if the move would be performed.

        Parameters
        ----------
        current_order : numpy.ndarray
            A 1 dimensional array that contains the order of the points to
            visit. All values are unique and are within the interval [0,size[.
            This is the current order.
        move : tuple of int
            Contains the move one wishes to know the effects on the quality of.

        Returns
        -------
        int or float
            The difference in quality if the move would be performed.

        """

        # get the changed distances
        # these are represented as a set of tuples of 2 ints that represent
        # the 2 unique indices between which the distance is changed.
        changed = self.changed_distances(self.eval_func._size, move)

        # init values
        next_solution_value = 0
        current_solution_value = 0

        # for all changed distances:
        # - add the original value to current_solution_value
        # - add the "changed" value to next_solution_value
        for distances in changed:

            # get indices
            frm = distances[0]
            to = distances[1]

            # add distance to current value
            current_solution_value += self.eval_func._distance_matrix[
                current_order[frm]][current_order[to]]

            # add distance to the "next" value

            # transform the indices so the indices return the value if the
            # move was performed
            (frm, to) = self.next_to_current(frm, to, move)

            next_solution_value += self.eval_func._distance_matrix[
                current_order[frm]][current_order[to]]

        return next_solution_value - current_solution_value


# functions for array_swap

def array_swap_changed_distances(size, move):
    """Aid function for delta evaluation.

    Works with the array_swap move type.
    This function returns the pairs who would have an altered evaluation
    value due to the move.

    Parameters
    ----------
    size : int
        The size of the array.
    move : tuple of int
        A tuple of 2 ints that represents a single unique move.

    Returns
    -------
    set of tuple
        this set contains a tuple with every (from,to) pair that would have
        an altered evaluation value due to the move.

    Examples
    --------
    Some simple examples to demonstrate the behaviour:

    .. doctest::

        >>> from lclpy.evaluation.deltaeval.delta_tsp \\
        ...     import array_swap_changed_distances \\
        ...         as changed_distances
        ... # init
        >>> size = 10
        ... # tests
        ... # since the order of the items in a set might be different,
        ... # they are compared to an equivalent set.
        >>> changed = changed_distances(size, (4, 8))
        >>> changed == {(3, 4), (4, 5), (7, 8), (8,9)}
        True
        >>> changed = changed_distances(size, (4, 9))
        >>> changed == {(3, 4), (4, 5), (8, 9), (9, 0)}
        True
        >>> changed = changed_distances(size, (0, 8))
        >>> changed == {(9, 0), (0, 1), (7, 8), (8, 9)}
        True
        >>> changed = changed_distances(size, (0, 9))
        >>> changed == {(0, 1), (8, 9), (9, 0)}
        True

    """

    changed_dist = set()

    # iterating over the 2 swapped indices
    for order_index in move:

        # get the change in the lower indices
        if order_index != 0:
            changed_dist.add((order_index - 1, order_index))
        else:

            # between index 0 and index _size-1, the pair is
            # (_size - 1, 0), this because we move from _size-1 to 0
            changed_dist.add((size - 1, 0))

        # get the change in the higher indices
        if order_index != size - 1:
            changed_dist.add((order_index, order_index + 1))
        else:
            changed_dist.add((size - 1, 0))

    return changed_dist


def array_swap_transform_next_index_to_current_index(frm, to, move):
    """Transforms frm and to depending on a move

    Works with the array_swap move type.
    This function transforms the indices frm and to so that they can
    be used as indices in the unaltered array, yet return the value
    they would have had if the move was actually performed and they
    were used as indices.

    Parameters
    ----------
    frm : int
        The from index that one wants to use in the array if the move was
        performed.
    to : int
        The to index that one wants to use in the array if the move was
        performed.
    move : tuple of int
        A tuple with that represents a single, unique move.

    Returns
    -------
    frm : int
        The index in the unaltered array that has the same value as the
        parameter frm in an array where the move was performed.
    to : int
        The index in the unaltered array that has the same value as the
        parameter to in an array where the move was performed.

    Examples
    --------
    Some simple examples, the indices remain the same, but the move
    changes:

    .. doctest::

        >>> from lclpy.evaluation.deltaeval.delta_tsp \\
        ...     import array_swap_transform_next_index_to_current_index \\
        ...         as transform_next_index_to_current_index
        >>> transform_next_index_to_current_index(1, 5, (1, 5))
        (5, 1)
        >>> transform_next_index_to_current_index(1, 5, (1, 3))
        (3, 5)
        >>> transform_next_index_to_current_index(1, 5, (0, 5))
        (1, 0)
        >>> transform_next_index_to_current_index(1, 5, (2, 3))
        (1, 5)


    """

    # transform frm so it returns the value that from would have if the
    # move was performed.
    if frm == move[0]:
        frm = move[1]
    elif frm == move[1]:
        frm = move[0]

    # transform to so it returns the value that from would have if the
    # move was performed.
    if to == move[0]:
        to = move[1]
    elif to == move[1]:
        to = move[0]

    return (frm, to)


# functions for array_reverse_order

def array_reverse_order_changed_distances(size, move):
    """Aid function for delta evaluation.

    Works with the array_reverse_order move type.
    This function returns the pairs who would have an altered evaluation
    value due to the move.

    Parameters
    ----------
    size : int
        The size of the array.
    move : tuple of int
        A tuple of 2 ints that represents a single valid move.

    Returns
    -------
    set of tuple
        this set contains a tuple with every (from,to) pair that would have
        an altered evaluation value due to the move.
        A pair (x, y) and a pair (y, x) are assumed to have different
        evaluation values.

    Examples
    --------
    Some simple examples to demonstrate the behaviour:

    .. doctest::

        >>> from lclpy.evaluation.deltaeval.delta_tsp import \\
        ...     array_reverse_order_changed_distances as \\
        ...         changed_distances
        ... # init
        >>> size = 10
        ... # tests
        ... # since the order of the items in a set might be different,
        ... # they are compared to an equivalent set.
        >>> changed = changed_distances(size, (4, 8))
        >>> changed == {(3, 4), (4, 5), (5, 6), (6, 7), (7, 8), (8, 9)}
        True
        >>> changed = changed_distances(size, (4, 9))
        >>> changed == {(3, 4), (4, 5), (5, 6),
        ...             (6, 7), (7, 8), (8, 9), (9, 0)}
        True
        >>> changed = changed_distances(size, (0, 4))
        >>> changed == {(9, 0), (0, 1), (1, 2), (2, 3), (3, 4), (4, 5)}
        True
        >>> changed = changed_distances(size, (0, 9))
        >>> changed == {(0, 1), (1, 2), (2, 3), (3, 4), (4, 5),
        ...             (5, 6), (6, 7), (7, 8), (8, 9), (9, 0)}
        True

    """

    changed_dist = set()

    # Calculating the distances that are always changed

    if (move[0] == 0):
        changed_dist.add((size - 1, 0))
    else:
        changed_dist.add((move[0] - 1, move[0]))

    if (move[1] == size - 1):
        changed_dist.add((size - 1, 0))
    else:
        changed_dist.add((move[1], move[1] + 1))

    # calculating the distances that are only changed if X -> Y causes a
    # different evaluation value than Y -> X

    for i in range(move[0], move[1]):

        changed_dist.add((i, i + 1))

    return changed_dist


def array_reverse_order_transform_next_index_to_current_index(frm, to, move):
    """Transforms frm and to depending on a move

    Works with the array_reverse_order move type.
    This function transforms the indices frm and to so that they can
    be used as indices in the unaltered array, yet return the value
    they would have had if the move was actually performed and they
    were used as indices.

    Parameters
    ----------
    frm : int
        The from index that one wants to use in the array if the move was
        performed.
    to : int
        The to index that one wants to use in the array if the move was
        performed.
    move : tuple of int
        A tuple with that represents a single, unique move.

    Returns
    -------
    frm : int
        The index in the unaltered array that has the same value as the
        parameter frm in an array where the move was performed.
    to : int
        The index in the unaltered array that has the same value as the
        parameter to in an array where the move was performed.

    Examples
    --------
    Some simple examples, the move remains the same, but the indices
    change:

    .. doctest::

        >>> from lclpy.evaluation.deltaeval.delta_tsp import \\
        ...     array_reverse_order_transform_next_index_to_current_index \\
        ...         as transform_next_index_to_current_index
        >>> transform_next_index_to_current_index(0, 10, (1, 8))
        (0, 10)
        >>> transform_next_index_to_current_index(0, 6, (1, 8))
        (0, 3)
        >>> transform_next_index_to_current_index(2, 3, (1, 8))
        (7, 6)
        >>> transform_next_index_to_current_index(1, 8, (1, 8))
        (8, 1)
        >>> transform_next_index_to_current_index(5, 10, (1, 8))
        (4, 10)


    """

    # check if the frm value is affected by the move
    if (frm >= move[0]) & (frm <= move[1]):

        # alter the value as necessary
        offset = frm - move[0]
        frm = move[1] - offset

    # check if the to value is affected by the move
    if (to >= move[0]) & (to <= move[1]):

        # alter the value as necessary
        offset = to - move[0]
        to = move[1] - offset

    return (frm, to)


# The method to return the other stuff.

def delta_tsp(eval_func, move_func):
    """Returns delta-eval class for a TSP problem.

    Note that if no methods for the problem can be found, that a placeholder
    method will be used. The placeholder methods will raise a
    NotImplementedError when called.

    Parameters
    ----------
    eval_func : AbstractEvaluationFunction
        The used evaluation function object.
    move_func : AbstractMove
        The used move object.

    Returns
    -------
    TSPDeltaEvaluate
        Class useable for delta evaluation of TSP problems.

    """

    move_type = move_func.get_move_type()

    if move_type is 'array_swap':
        return TSPDeltaEvaluate(eval_func, array_swap_changed_distances,
                                array_swap_transform_next_index_to_current_index)
    if move_type is 'array_reverse_order':
        return TSPDeltaEvaluate(eval_func,
                                array_reverse_order_changed_distances,
                                array_reverse_order_transform_next_index_to_current_index)
    else:
        raise NotImplementedError



"""This module contains several functions to get statistical data from the
results of the function benchmark.
"""

from operator import itemgetter
import numpy
import statistics
from collections import namedtuple


def _func_on_best_values(benchmark_result, func):
    """The func will be performed on the list of best values.

    Parameters
    ----------
    benchmark_result : list of list of list of namedtuple
        The result from a benchmark.
    func
        The function that will be performed on all namedtuples of an
        algorithm-problem pair.

    Returns
    -------
    numpy.ndarray
        The 2 dimensional ndarray will contain the value that the function
        returns. Note that the indices of a certain algorithm-problem pair in
        the benchmark_result will be the same as the indices one needs to get
        the results for that pair.

    """

    position_getter = itemgetter(1)

    result = []

    for algorithm in benchmark_result:

        aid_list = []

        for problem in algorithm:

            aid_list.append(func(map(position_getter, problem)))

        result.append(aid_list)

    return numpy.array(result)


# Functions that do something with the best_value under here

def mean(benchmark_result):
    """A function to calculate the mean of the best_values.

    Parameters
    ----------
    benchmark_result : list of list of list of namedtuple
        The result from a benchmark.

    Returns
    -------
    numpy.ndarray
        A 2D array containing the mean of the best_value for every
        algorithm-problem pair. Note that the indices of a certain
        algorithm-problem pair in the benchmark_result will be the same as the
        indices one needs to get the results for that pair.

    """

    return _func_on_best_values(benchmark_result, statistics.mean)


def median(benchmark_result):
    """A function to calculate the median of the best_values.

    Parameters
    ----------
    benchmark_result : list of list of list of namedtuple
        The result from a benchmark.

    Returns
    -------
    numpy.ndarray
        A 2D array containing the median of the best_value for every
        algorithm-problem pair. Note that the indices of a certain
        algorithm-problem pair in the benchmark_result will be the same as the
        indices one needs to get the results for that pair.

    """

    return _func_on_best_values(benchmark_result, statistics.median)


def stdev(benchmark_result):
    """A function to calculate the standard deviation of the best_values.

    Parameters
    ----------
    benchmark_result : list of list of list of namedtuple
        The result from a benchmark.

    Returns
    -------
    numpy.ndarray
        A 2D array containing the median of the standard variation for every
        algorithm-problem pair. Note that the indices of a certain
        algorithm-problem pair in the benchmark_result will be the same as the
        indices one needs to get the results for that pair.

    """

    return _func_on_best_values(benchmark_result, statistics.stdev)


def biggest(benchmark_result):
    """A function to get the biggest best_values.

    Parameters
    ----------
    benchmark_result : list of list of list of namedtuple
        The result from a benchmark.

    Returns
    -------
    numpy.ndarray
        A 2D array containing the biggest of the best_value for every
        algorithm-problem pair. Note that the indices of a certain
        algorithm-problem pair in the benchmark_result will be the same as the
        indices one needs to get the results for that pair.

    """

    return _func_on_best_values(benchmark_result, max)


def smallest(benchmark_result):
    """A function to get the smallest best_values.

    Parameters
    ----------
    benchmark_result : list of list of list of namedtuple
        The result from a benchmark.

    Returns
    -------
    numpy.ndarray
        A 2D array containing the smallest of the best_value for every
        algorithm-problem pair. Note that the indices of a certain
        algorithm-problem pair in the benchmark_result will be the same as the
        indices one needs to get the results for that pair.

    """

    return _func_on_best_values(benchmark_result, min)


# Functions that do something with the data collected during the run under here


def _func_on_data(benchmark_result, func, position):
    """The func will be performed on the list of best values.

    Parameters
    ----------
    benchmark_result : list of list of list of namedtuple
        The result from a benchmark.
    func
        The function that will be performed on all namedtuples of an
        algorithm-problem pair.
    position : int
        The position of the value in the namedtuples that the function will
        use.

    Returns
    -------
    numpy.ndarray
        The 2 dimensional ndarray will contain the value that the function
        returns. Note that the indices of a certain algorithm-problem pair in
        the benchmark_result will be the same as the indices one needs to get
        the results for that pair.

    """

    result = []

    for algorithm in benchmark_result:

        aid_list = []

        for problem in algorithm:

            last_items = []

            for run in problem:
                last_items.append(run.data[position][-1])

            aid_list.append(func(last_items))

        result.append(aid_list)

    return numpy.array(result)


# time

def time_mean(benchmark_result):
    """A function to calculate the mean of the last time-point.

    Parameters
    ----------
    benchmark_result : list of list of list of namedtuple
        The result from a benchmark.

    Returns
    -------
    numpy.ndarray
        A 2D array containing the mean of the last time-point for every
        algorithm-problem pair. Note that the indices of a certain
        algorithm-problem pair in the benchmark_result will be the same as the
        indices one needs to get the results for that pair.

    """

    return _func_on_data(benchmark_result, statistics.mean, 0)


def time_median(benchmark_result):
    """A function to calculate the median of the last time-point.

    Parameters
    ----------
    benchmark_result : list of list of list of namedtuple
        The result from a benchmark.

    Returns
    -------
    numpy.ndarray
        A 2D array containing the median of the last time-point for every
        algorithm-problem pair. Note that the indices of a certain
        algorithm-problem pair in the benchmark_result will be the same as the
        indices one needs to get the results for that pair.

    """

    return _func_on_data(benchmark_result, statistics.median, 0)


def time_stdev(benchmark_result):
    """A function to calculate the standard deviation of the last time-point.

    Parameters
    ----------
    benchmark_result : list of list of list of namedtuple
        The result from a benchmark.

    Returns
    -------
    numpy.ndarray
        A 2D array containing the standard deviation of the last time-point for
        every algorithm-problem pair. Note that the indices of a certain
        algorithm-problem pair in the benchmark_result will be the same as the
        indices one needs to get the results for that pair.

    """

    return _func_on_data(benchmark_result, statistics.stdev, 0)


def time_max(benchmark_result):
    """A function to get the longest execution time.

    Parameters
    ----------
    benchmark_result : list of list of list of namedtuple
        The result from a benchmark.

    Returns
    -------
    numpy.ndarray
        A 2D array containing the longest time for every algorithm-problem
        pair. Note that the indices of a certain algorithm-problem pair in the
        benchmark_result will be the same as the indices one needs to get the
        results for that pair.

    """

    return _func_on_data(benchmark_result, max, 0)


def time_min(benchmark_result):
    """A function to get the shortest execution time.

    Parameters
    ----------
    benchmark_result : list of list of list of namedtuple
        The result from a benchmark.

    Returns
    -------
    numpy.ndarray
        A 2D array containing the shortest time for every algorithm-problem
        pair. Note that the indices of a certain algorithm-problem pair in the
        benchmark_result will be the same as the indices one needs to get the
        results for that pair.

    """

    return _func_on_data(benchmark_result, min, 0)


# iterations

def iterations_mean(benchmark_result):
    """A function to calculate the mean of the amount of iterations.

    Parameters
    ----------
    benchmark_result : list of list of list of namedtuple
        The result from a benchmark.

    Returns
    -------
    numpy.ndarray
        A 2D array containing the mean of the amount of iterations for every
        algorithm-problem pair. Note that the indices of a certain
        algorithm-problem pair in the benchmark_result will be the same as the
        indices one needs to get the results for that pair.

    """

    return _func_on_data(benchmark_result, statistics.mean, 1)


def iterations_median(benchmark_result):
    """A function to calculate the median of the amount of iterations.

    Parameters
    ----------
    benchmark_result : list of list of list of namedtuple
        The result from a benchmark.

    Returns
    -------
    numpy.ndarray
        A 2D array containing the median of the amount of iterations for every
        algorithm-problem pair. Note that the indices of a certain
        algorithm-problem pair in the benchmark_result will be the same as the
        indices one needs to get the results for that pair.

    """

    return _func_on_data(benchmark_result, statistics.median, 1)


def iterations_stdev(benchmark_result):
    """A function to calculate the standard deviation of the amount of iterations.

    Parameters
    ----------
    benchmark_result : list of list of list of namedtuple
        The result from a benchmark.

    Returns
    -------
    numpy.ndarray
        A 2D array containing the standard deviation of the amount of
        iterations for every algorithm-problem pair. Note that the indices of a
        certain algorithm-problem pair in the benchmark_result will be the same
        as the indices one needs to get the results for that pair.

    """

    return _func_on_data(benchmark_result, statistics.stdev, 1)


def iterations_max(benchmark_result):
    """A function to get the most iterations.

    Parameters
    ----------
    benchmark_result : list of list of list of namedtuple
        The result from a benchmark.

    Returns
    -------
    numpy.ndarray
        A 2D array containing the most iterations for every algorithm-problem
        pair. Note that the indices of a certain algorithm-problem pair in the
        benchmark_result will be the same as the indices one needs to get the
        results for that pair.

    """

    return _func_on_data(benchmark_result, max, 1)


def iterations_min(benchmark_result):
    """A function to get the least iterations.

    Parameters
    ----------
    benchmark_result : list of list of list of namedtuple
        The result from a benchmark.

    Returns
    -------
    numpy.ndarray
        A 2D array containing the least iterations for every algorithm-problem
        pair. Note that the indices of a certain algorithm-problem pair in the
        benchmark_result will be the same as the indices one needs to get the
        results for that pair.

    """

    return _func_on_data(benchmark_result, min, 1)


def stat(benchmark_result, algorithm_names=None, problem_names=None):
    """A function to get some common characteristics from a benchmark.

    Parameters
    ----------
    benchmark_result : list of list of list of namedtuple
        The result from a benchmark.
    print : bool, optional
        If True, the results will be printed to the command line.
        If False the results will not be printed to the command line.

    Returns
    -------
    namedtuple of namedtuple of numpy.ndarray
        The result is divided in 3 main parts: best_value, time and iterations.
        Every main part contains the results of different statistics:

        - mean
        - median
        - stdev
        - max
        - min

        The result for each of those are kept in a 2D array. Note that the
        indices of a certain algorithm-problem pair in the benchmark_result
        will be the same as the indices one needs to get the results for that
        pair.

    """

    # get the data
    # best_value

    best_value_tuple_class = namedtuple('best_value_statistics', (
        'mean', 'median', 'stdev', 'max', 'min'))

    _mean = mean(benchmark_result)

    _median = median(benchmark_result)

    _stdev = stdev(benchmark_result)

    _biggest = biggest(benchmark_result)

    _smallest = smallest(benchmark_result)

    best_value_tuple = best_value_tuple_class(
        _mean, _median, _stdev, _biggest, _smallest)

    # time

    time_tuple_class = namedtuple('time_statistics', (
        'mean', 'median', 'stdev', 'max', 'min'))

    _time_mean = time_mean(benchmark_result)

    _time_median = time_median(benchmark_result)

    _time_stdev = time_stdev(benchmark_result)

    _time_max = time_max(benchmark_result)

    _time_min = time_min(benchmark_result)

    time_tuple = time_tuple_class(
        _time_mean, _time_median, _time_stdev, _time_max, _time_min)

    # iterations

    iterations_tuple_class = namedtuple('iterations_statistics', (
        'mean', 'median', 'stdev', 'max', 'min'))

    _iterations_mean = iterations_mean(benchmark_result)

    _iterations_median = iterations_median(benchmark_result)

    _iterations_stdev = iterations_stdev(benchmark_result)

    _iterations_max = iterations_max(benchmark_result)

    _iterations_min = iterations_min(benchmark_result)

    iterations_tuple = iterations_tuple_class(
        _iterations_mean, _iterations_median,
        _iterations_stdev, _iterations_max, _iterations_min)

    # combine namedtuples

    combined_tuple_class = namedtuple(
        'benchmark_statistics', ('best_value', 'time', 'iterations'))

    return combined_tuple_class(best_value_tuple, time_tuple, iterations_tuple)



from lclpy.localsearch.abstract_local_search import AbstractLocalSearch

from lclpy.aidfunc.is_improvement_func import bigger, smaller
from lclpy.aidfunc.pass_func import pass_func
from lclpy.aidfunc.add_to_data_func import add_to_data_func
from lclpy.aidfunc.convert_data import convert_data
from lclpy.aidfunc.logging import log_improvement

from lclpy.localsearch.tabusearch.tabu_list import TabuList

from collections import namedtuple


class TabuSearch(AbstractLocalSearch):
    """Performs a tabu search on the given problem.

    Parameters
    ----------
    problem : AbstractLocalSearchProblem
        Contains all the data needed for the specific problem.
    termination_criterion : AbstractTerminationCriterion
        Implements a termination criterion to terminate the algorithm.
    diff_state_func : AbstractDiffState
        A class that implement a diff_state_func for tabu search.
    list_size : int, optional
        The size of the tabu list. The default is 7.
    minimise : bool, optional
        Will minimise if this parameter is True, maximise if it is False.
        The default is True.
    benchmarking : bool, optional
        Should be True if one wishes benchmarks to be kept, should be False if
        one wishes no benchmarks to be made. Default is False.
    logging: bool, optional
        Improvements will be logged to the command line if this variable is
        True. Default is True.

    Attributes
    ----------
    _problem : AbstractLocalSearchProblem
        Contains all the data needed for the specific problem.
    _termination_criterion : AbstractTerminationCriterion
        Implements a termination criterion to terminate the algorithm.
    _diff : method
        A method of a AbstractDiffState class to indicate how a move alters the
        state
    _list_size : int
        Size of the tabu list.
    _tabu_list : TabuList
        The used tabu list.
    _is_better
        Function used to determine if a certain value is an improvement.
    _minimise : bool
        Variable that indicates if the function is maximising or minimising.
    _best_found_delta_base_value : float
        Initialisation value for the delta value of each iteration. It's
        infinite when minimising or minus infinite when maximising.
    data : list of tuple
        Data useable for benchmarking. Will be None if no benchmarks are made.
    _data_append
        Function to append new data-points to data. Will do nothing if no
        benchmarks are made.
    _log_improvement
        Function to write logs to the command line. Will do nothing if no logs
        are made.

    Examples
    --------
    An example of minimising:

    .. doctest::

        >>> import numpy
        >>> import random
        >>> from lclpy.localsearch.tabusearch.tabu_search import TabuSearch
        >>> from lclpy.localsearch.move.tsp_array_swap import TspArraySwap
        >>> from lclpy.evaluation.tsp_evaluation_function \\
        ...     import TspEvaluationFunction
        >>> from lclpy.termination.max_seconds_termination_criterion \\
        ...     import MaxSecondsTerminationCriterion
        >>> from lclpy.localsearch.tabusearch.sum_diff_state \\
        ...     import SumDiffState
        >>> from lclpy.problem.array_problem import ArrayProblem
        ... # seed random
        ... # (used here to always get the same output, this obviously is not
        ... #                                  needed in your implementation.)
        >>> random.seed(0)
        ... # init problem
        >>> distance_matrix = numpy.array(
        ... [[0, 2, 5, 8],
        ...  [2, 0, 4, 1],
        ...  [5, 4, 0, 7],
        ...  [8, 1, 7, 0]])
        >>> size = distance_matrix.shape[0]
        >>> move = TspArraySwap(size)
        >>> evaluation = TspEvaluationFunction(distance_matrix, move)
        >>> problem = ArrayProblem(evaluation, move, size)
        ... # init termination criterion
        >>> termination = MaxSecondsTerminationCriterion(10)
        ... # init diff_state_func
        >>> diff_state_func = SumDiffState()
        ... # init TabuSearch
        >>> tabu_search = TabuSearch(problem, termination, diff_state_func, 5, logging=False)
        ... # run algorithm
        >>> tabu_search.run()
        Results(best_order=array([0, 1, 3, 2]), best_value=15, data=None)

    An example of maximising, note that the distance matrix is different:

    .. doctest::

        >>> import numpy
        >>> import random
        >>> from lclpy.localsearch.tabusearch.tabu_search import TabuSearch
        >>> from lclpy.localsearch.move.tsp_array_swap import TspArraySwap
        >>> from lclpy.evaluation.tsp_evaluation_function \\
        ...     import TspEvaluationFunction
        >>> from lclpy.termination.max_seconds_termination_criterion \\
        ...     import MaxSecondsTerminationCriterion
        >>> from lclpy.localsearch.tabusearch.sum_diff_state \\
        ...     import SumDiffState
        >>> from lclpy.problem.array_problem import ArrayProblem
        ... # seed random
        ... # (used here to always get the same output, this obviously is not
        ... #                                  needed in your implementation.)
        >>> random.seed(0)
        ... # init problem
        >>> distance_matrix = numpy.array(
        ... [[0, 8, 5, 2],
        ...  [8, 0, 4, 7],
        ...  [5, 4, 0, 1],
        ...  [2, 7, 1, 0]])
        >>> size = distance_matrix.shape[0]
        >>> move = TspArraySwap(size)
        >>> evaluation = TspEvaluationFunction(distance_matrix, move)
        >>> problem = ArrayProblem(evaluation, move, size)
        ... # init termination criterion
        >>> termination = MaxSecondsTerminationCriterion(10)
        ... # init diff_state_func
        >>> diff_state_func = SumDiffState()
        ... # init TabuSearch
        >>> tabu_search = TabuSearch(problem, termination, diff_state_func, 5,
        ...                          False, logging=False)
        ... # run algorithm
        >>> tabu_search.run()
        Results(best_order=array([0, 1, 3, 2]), best_value=21, data=None)


    """

    def __init__(self, problem, termination_criterion, diff_state_func,
                 list_size=7, minimise=True, benchmarking=False, logging=True):

        super().__init__()

        self._problem = problem
        self._termination_criterion = termination_criterion
        self._diff = diff_state_func.diff
        self._list_size = list_size
        self._tabu_list = TabuList(list_size)
        self._minimise = minimise

        if minimise:
            self._is_better = smaller
            self._best_found_delta_base_value = float("inf")
        else:
            self._is_better = bigger
            self._best_found_delta_base_value = float("-inf")

        if benchmarking:
            self.data = []
            self._data_append = add_to_data_func
        else:
            self.data = None
            self._data_append = pass_func

        if logging:
            self._log_improvement = log_improvement
        else:
            self._log_improvement = pass_func

    def run(self):
        """Starts running the tabu search.

        Returns
        -------
        best_order : numpy.ndarray
            The best found order.
        best_value : int or float
            The evaluation value of the best found order.
        data : None or collections.namedtuple
            Data useable for benchmarking. If no benchmarks were made, it will
            be None. The namedtuple contains the following data in
            numpy.ndarrays:
            time, iteration, value, best_value


        """

        # init
        base_value = self._problem.evaluate()
        self._problem.set_as_best(base_value)

        # init iteration (used to count the amount of iterations)
        iteration = 0

        # add to data
        self._data_append(self.data, iteration, base_value, base_value)

        # init termination criterion
        self._termination_criterion.check_first_value(base_value)
        self._termination_criterion.start_timing()

        # main loop
        while self._termination_criterion.keep_running():

            # search the neighbourhood for the best move
            best_found_delta = self._best_found_delta_base_value
            best_found_move = None

            for move in self._problem.get_moves():

                # check quality move
                delta = self._problem.evaluate_move(move)

                # checks how the move alters the current state
                diff = self._diff(move)

                # if not in tabu list --> not similar to earlier performed
                # moves --> if delta better than old best move
                # --> becomes the best move

                if not self._tabu_list.contains(diff) and \
                        self._is_better(best_found_delta, delta):
                    best_found_delta = delta
                    best_found_move = move
                    best_found_diff = diff

            # the best found move will be used as the next move
            # alter state problem
            base_value = base_value + best_found_delta

            # check if a move was found
            if best_found_move is not None:

                self._problem.move(best_found_move)

                # if better than best found --> new best_found
                if self._is_better(self._problem.best_order_value,
                                   base_value):
                    self._problem.set_as_best(base_value)

                # add diff to tabu list
                self._tabu_list.add(best_found_diff)

                # log if needed
                self._log_improvement(base_value)

                # add to data
                self._data_append(self.data, iteration,
                                  base_value, self._problem.best_order_value)

                self._termination_criterion.check_new_value(base_value)

                # functions _termination_criterion called
                self._termination_criterion.check_new_value(base_value)

            else:
                # no move found --> we're stuck --> break loop
                break

            iteration += 1
            self._termination_criterion.iteration_done()

        # last data point
        self._data_append(self.data, iteration, base_value,
                          self._problem.best_order_value)

        # if we have data:
        # convert data to something easier to plot
        if self.data is not None:

            # convert to tuple of list
            data = convert_data(self.data)

            # make namedtuple
            DataAsLists = namedtuple(
                'Data', ['time', 'iteration', 'value', 'best_value'])

            data = DataAsLists(data[0], data[1], data[2], data[3])

        else:
            data = None

        # return results

        Results = namedtuple('Results', ['best_order', 'best_value', 'data'])

        return Results(self._problem.best_order,
                       self._problem.best_order_value,
                       data)

    def reset(self):
        """Resets the object back to it's state after init.

        Raises
        ------
        NotImplementedError
            If the problem or termination criterion have no reset method
            implemented.

        Examples
        --------
        An example of minimising, reset after run:

        .. doctest::

            >>> import numpy
            >>> import random
            >>> from lclpy.localsearch.tabusearch.tabu_search \\
            ...     import TabuSearch
            >>> from lclpy.localsearch.move.tsp_array_swap \\
            ...     import TspArraySwap
            >>> from lclpy.evaluation.tsp_evaluation_function \\
            ...     import TspEvaluationFunction
            >>> from lclpy.termination.max_iterations_termination_criterion \\
            ...     import MaxIterationsTerminationCriterion
            >>> from lclpy.localsearch.tabusearch.sum_diff_state \\
            ...     import SumDiffState
            >>> from lclpy.problem.array_problem import ArrayProblem
            ... # seed random
            ... # (used here to always get the same output, this obviously is
            ... #                           not needed in your implementation.)
            >>> random.seed(0)
            ... # init problem
            >>> distance_matrix = numpy.array(
            ... [[0, 2, 5, 8],
            ...  [2, 0, 4, 1],
            ...  [5, 4, 0, 7],
            ...  [8, 1, 7, 0]])
            >>> size = distance_matrix.shape[0]
            >>> move = TspArraySwap(size)
            >>> evaluation = TspEvaluationFunction(distance_matrix, move)
            >>> problem = ArrayProblem(evaluation, move, size)
            ... # init termination criterion
            >>> termination = MaxIterationsTerminationCriterion(5)
            ... # init TabuSearch
            ... # init diff_state_func
            >>> diff_state_func = SumDiffState()
            ... # init TabuSearch
            >>> tabu_search = TabuSearch(problem, termination, diff_state_func,
            ...                          5, True, logging=False)
            ... # state before running
            >>> tabu_search._problem._order
            array([0, 1, 2, 3])
            >>> tabu_search._termination_criterion.keep_running()
            True
            >>> len(tabu_search._tabu_list._list)
            0
            >>> # run algorithm
            >>> tabu_search.run()
            Results(best_order=array([0, 1, 3, 2]), best_value=15, data=None)
            >>> # tests reset
            >>> # before reset
            >>> tabu_search._problem._order
            array([0, 3, 2, 1])
            >>> tabu_search._termination_criterion.keep_running()
            True
            >>> len(tabu_search._tabu_list._list)
            3
            >>> # reset
            >>> tabu_search.reset()
            ... # after reset
            >>> tabu_search._problem._order
            array([0, 1, 2, 3])
            >>> tabu_search._termination_criterion.keep_running()
            True
            >>> len(tabu_search._tabu_list._list)
            0


        """

        self._problem.reset()
        self._termination_criterion.reset()

        self._tabu_list = TabuList(self._list_size)

        if self.data is not None:
            self.data = []



from abc import ABC, abstractmethod


class AbstractEvaluationFunction(ABC):
    """A template to create evaluation functions.

    Evaluationfunctions are used to determine the quality of a problem.

    """

    def __init__(self):
        super().__init__()

    def get_problem_type(self):
        """Returns the problem type.

        Returns
        -------
        str
            The problem type.

        """
        raise NotImplementedError

    @abstractmethod
    def evaluate(self, current_data):
        """Evaluates current_solution

        Parameters
        ----------
        current_data
            A datastructure that contains information about the current state
            of an AbstractLocalSearchSolution.

        Returns
        -------
        int or float
            an indication of the quality of current_data

        """
        pass

    def delta_evaluate(self, current_data, move):
        """Evaluates the difference in quality between two solutions.

        The two compared solutions are the current problem and the problem
        if the move was performed. The move is not actually performed.

        For this function to work, get_problem_type needs to be implemented,
        the move-object(s) need to have get_move_type implemented and the
        delta evaluate methods need to be properly implemented in the deltaeval
        package.

        Parameters
        ----------
        current_data
            A datastructure that contains information about the current state
            of an AbstractLocalSearchSolution.

        move
            Represents a move. The potential result of this move is evaluated
            against the current state (current_data contains information about
            the current state.).


        Returns
        -------
        int or float
            the difference in quality between the current state and the
            potential next state.

        """
        raise NotImplementedError



from abc import ABC, abstractmethod


class AbstractCoolingFunction(ABC):
    """A template class to implement a cooling function for simulated annealing.

    Implementations of this class are meant to be used in a simulated annealing
    algorithm. Implementations are supposed to implement the cooling function.

    """

    def __init__(self):
        super().__init__()

    @abstractmethod
    def next_temperature(self, old_temperature):
        """Function to get the next temperature

        Parameters
        ----------
        old_temperature : int or float
            The old temperature

        Returns
        -------
        int or float
            The new temperature

        """

        pass



from lclpy.localsearch.simulatedannealing.abstr_iterations_temp_function \
    import AbstrIterationsTempFunction


class CnstIterationsTempFunction(AbstrIterationsTempFunction):
    """Class for determining the number of iterations for a temperature.

    This class is meant to be used in a simulated annealing algorithm.
    The amount of iterations is constant and is not influenced by the
    temperature.

    Parameters
    ----------
    iterations : int
        The amount of iterations for all temperatures.

    Attributes
    ----------
    _iterations : int
        The amount of iterations for all temperatures. Default is 1000.

    Examples
    --------
    Default behaviour:

    .. doctest::

        >>> from lclpy.localsearch.simulatedannealing.cnst_iterations_temp_function \\
        ...     import CnstIterationsTempFunction
        ... # init
        >>> test = CnstIterationsTempFunction()
        ... # tests
        >>> test.get_iterations(1000)
        1000
        >>> test.get_iterations(359)
        1000
        >>> test.get_iterations(0)
        1000

    With 800 iterations:

    .. doctest::

        >>> from lclpy.localsearch.simulatedannealing.cnst_iterations_temp_function \\
        ...     import CnstIterationsTempFunction
        ... # init
        >>> test = CnstIterationsTempFunction(800)
        ... # tests
        >>> test.get_iterations(1000)
        800
        >>> test.get_iterations(359)
        800
        >>> test.get_iterations(0)
        800

    """

    def __init__(self, iterations=1000):
        super().__init__()
        self._iterations = iterations

    def get_iterations(self, temperature):
        """Returns the amount of iterations for a certain temperature.

        Parameters
        ----------
        temperature : int or float
            The "temperature" in the simulated annealing algorithm.

        Returns
        -------
        int
            The amount of iterations for the temperature. (Here always
            _iterations)

        """

        return self._iterations



from abc import ABC, abstractmethod


class AbstractDiffState(ABC):
    """A template class to implement a diff_state_func for tabu search.

    Implementations of this class are meant to be used in a tabu search
    algorithm. Implementations are supposed to implement the diff function.
    A specific implementation might be needed for different move types.

    """

    def __init__(self):
        super().__init__()

    @abstractmethod
    def diff(self, move):
        """Indicates how the move would alter the state if it was performed.

        Parameters
        ----------
        move : tuple of int
            A representation of a move.

        Returns
        -------
        int
            An indication of how the move alters the state.

        """

        pass



from lclpy.localsearch.tabusearch.abstract_diff_state import AbstractDiffState


class SumDiffState(AbstractDiffState):
    """A class that implement a diff_state_func for tabu search.

    Examples
    --------
    A simple example:

    .. doctest::

        >>> from lclpy.localsearch.tabusearch.sum_diff_state \\
        ... 	import SumDiffState
        ... # init
        >>> test = SumDiffState()
        ... # tests
        >>> test.diff((12, 24))
        36
        >>> test.diff((3, 8))
        11
        >>> test.diff((50, 70))
        120

    """

    def __init__(self):
        super().__init__()

    def diff(self, move):
        """Indicates how the move would alter the state if it was performed.

        Parameters
        ----------
        move : tuple of int
            A tuple of 2 ints.

        Returns
        -------
        int
            The sum of the values of the tuple.

        """

        return move[0] + move[1]



from lclpy.localsearch.abstract_local_search import AbstractLocalSearch

from lclpy.localsearch.acceptance.simulated_annealing_acceptance_function \
    import SimulatedAnnealingAcceptanceFunction

from lclpy.aidfunc.is_improvement_func \
    import bigger, bigger_or_equal, smaller, smaller_or_equal
from lclpy.aidfunc.pass_func import pass_func
from lclpy.aidfunc.add_to_data_func import add_to_data_func
from lclpy.aidfunc.convert_data import convert_data
from lclpy.aidfunc.logging import log_improvement, log_passed_worse

from collections import namedtuple


class SimulatedAnnealing(AbstractLocalSearch):
    """Performs a simulated annealing algorithm with the given parameters.

    Parameters
    ----------
    problem : AbstractLocalSearchProblem
        Contains all the data needed for the specific problem.
    termination_criterion : AbstractTerminationCriterion
        Implements a termination criterion to terminate the algorithm.
    cooling_function : AbstractCoolingFunction
        Implements the cooling function for the algorithm.
    iterations_for_temp_f : AbstrIterationsTempFunction
        Implements a function to determine the amount of iterations for a
        certain temperature.
    start_temperature : int, optional
        The starting temperature for the simulated annealing. The default is
        2000.
    minimise : bool, optional
        Will minimise if this parameter is True, maximise if it is False.
        The default is True.
    benchmarking : bool, optional
        Should be True if one wishes benchmarks to be kept, should be False if
        one wishes no benchmarks to be made. Default is False.
    logging: bool, optional
        Improvements and passed worse solutions will be logged to the command
        line if this variable is True. Default is True.

    Attributes
    ----------
    _problem : AbstractLocalSearchProblem
        Contains all the data needed for the specific problem.
    _termination_criterion : AbstractTerminationCriterion
        Implements a termination criterion to terminate the algorithm.
    _cooling_function : AbstractCoolingFunction
        Implements the cooling function for the algorithm.
    _iterations_for_temp_f : AbstrIterationsTempFunction
        Implements a function to determine the amount of iterations for a
        certain temperature.
    _start_temperature : int
        The start "temperature".
    _temperature : int
        The current "temperature".
    _acceptance_function : SimulatedAnnealingAcceptanceFunction
        The acceptance function used.
    _is_improvement
        Function used to determine if a certain delta value is an improvement
        or equal to the current value.
    _is_better
        Function used to determine if a certain value is an improvement.
    data : list of tuple
        Data useable for benchmarking. Will be None if no benchmarks are made.
    _data_append
        Function to append new data-points to data. Will do nothing if no
        benchmarks are made.
    _log_improvement
        Function to write logs to the command line. Will do nothing if no logs
        are made.
    _log_passed_worse
        Function to write logs to the command line. Will do nothing if no logs
        are made.

    Examples
    --------
    Minimising example:

    .. doctest::

        >>> import numpy
        >>> import random
        >>> from lclpy.localsearch.move.tsp_array_swap import TspArraySwap
        >>> from lclpy.localsearch.simulatedannealing.simulated_annealing \\
        ...     import SimulatedAnnealing
        >>> from lclpy.localsearch.simulatedannealing.geometric_cooling_function \\
        ...     import GeometricCoolingFunction
        >>> from lclpy.localsearch.simulatedannealing.cnst_iterations_temp_function \\
        ...     import CnstIterationsTempFunction
        >>> from lclpy.evaluation.tsp_evaluation_function \\
        ...     import TspEvaluationFunction
        >>> from lclpy.termination.min_temperature_termination_criterion \\
        ...     import MinTemperatureTerminationCriterion
        >>> from lclpy.problem.array_problem import ArrayProblem
        ... # seed random
        ... # (used here to always get the same output, this obviously is not
        ... #                                  needed in your implementation.)
        >>> random.seed(0)
        ... # init problem
        >>> distance_matrix = numpy.array(
        ... [[0, 2, 5, 8],
        ...  [2, 0, 4, 1],
        ...  [5, 4, 0, 7],
        ...  [8, 1, 7, 0]])
        >>> size = distance_matrix.shape[0]
        >>> move = TspArraySwap(size)
        >>> evaluation = TspEvaluationFunction(distance_matrix, move)
        >>> problem = ArrayProblem(evaluation, move, size)
        ... # init termination criterion
        >>> termination_criterion = MinTemperatureTerminationCriterion()
        ... # init cooling function
        >>> cooling_func = GeometricCoolingFunction()
        ... # init CnstIterationsTempFunction
        ... # (determines amount of itertions in function of the temperature)
        >>> i_for_temp = CnstIterationsTempFunction()
        ... # init SimulatedAnnealing
        >>> algorithm = SimulatedAnnealing(problem, termination_criterion,
        ...                                cooling_func, i_for_temp, logging=False)
        ... # run algorithm
        >>> algorithm.run()
        Results(best_order=array([0, 2, 3, 1]), best_value=15, data=None)

    Maximising example, note that the distance matrix is different:

    .. doctest::

        >>> import numpy
        >>> import random
        >>> from lclpy.localsearch.move.tsp_array_swap import TspArraySwap
        >>> from lclpy.localsearch.simulatedannealing.simulated_annealing \\
        ...     import SimulatedAnnealing
        >>> from lclpy.localsearch.simulatedannealing.geometric_cooling_function \\
        ...     import GeometricCoolingFunction
        >>> from lclpy.localsearch.simulatedannealing.cnst_iterations_temp_function \\
        ...     import CnstIterationsTempFunction
        >>> from lclpy.evaluation.tsp_evaluation_function \\
        ...     import TspEvaluationFunction
        >>> from lclpy.termination.min_temperature_termination_criterion \\
        ...     import MinTemperatureTerminationCriterion
        >>> from lclpy.problem.array_problem import ArrayProblem
        ... # seed random
        ... # (used here to always get the same output, this obviously is not
        ... #                                  needed in your implementation.)
        >>> random.seed(0)
        ... # init problem
        >>> distance_matrix = numpy.array(
        ... [[0, 8, 5, 2],
        ...  [8, 0, 4, 7],
        ...  [5, 4, 0, 1],
        ...  [2, 7, 1, 0]])
        >>> size = distance_matrix.shape[0]
        >>> move = TspArraySwap(size)
        >>> evaluation = TspEvaluationFunction(distance_matrix, move)
        >>> problem = ArrayProblem(evaluation, move, size)
        ... # init termination criterion
        >>> termination_criterion = MinTemperatureTerminationCriterion()
        ... # init cooling function
        >>> cooling_func = GeometricCoolingFunction()
        ... # init CnstIterationsTempFunction
        ... # (determines amount of itertions in function of the temperature)
        >>> i_for_temp = CnstIterationsTempFunction()
        ... # init SimulatedAnnealing
        >>> algorithm = SimulatedAnnealing(
        ...     problem, termination_criterion,
        ...     cooling_func, i_for_temp, minimise=False, logging=False)
        ... # run algorithm
        >>> algorithm.run()
        Results(best_order=array([0, 2, 3, 1]), best_value=21, data=None)

    """

    def __init__(self, problem, termination_criterion,
                 cooling_function, iterations_for_temp_f,
                 start_temperature=2000, minimise=True,
                 benchmarking=False, logging=True):

        super().__init__()

        self._problem = problem
        self._termination_criterion = termination_criterion
        self._cooling_function = cooling_function
        self._iterations_for_temp_f = iterations_for_temp_f
        self._start_temperature = start_temperature
        self._temperature = start_temperature
        self._acceptance_function = SimulatedAnnealingAcceptanceFunction()

        if minimise:
            self._is_improvement = smaller_or_equal
            self._is_better = smaller
            self._acceptance_function = SimulatedAnnealingAcceptanceFunction()
        else:
            self._is_improvement = bigger_or_equal
            self._is_better = bigger
            self._acceptance_function = \
                SimulatedAnnealingAcceptanceFunction(diff_multiplier=-1)

        if benchmarking:
            self.data = []
            self._data_append = add_to_data_func
        else:
            self.data = None
            self._data_append = pass_func

        if logging:
            self._log_improvement = log_improvement
            self._log_passed_worse = log_passed_worse
        else:
            self._log_improvement = pass_func
            self._log_passed_worse = pass_func

    def run(self):
        """Starts running the simulated annealing algorithm.

        Returns
        -------
        best_order : numpy.ndarray
            The best found order.
        best_value : int or float
            The evaluation value of the best found order.
        data : None or collections.namedtuple
            Data useable for benchmarking. If no benchmarks were made, it will
            be None. The namedtuple contains the following data in
            numpy.ndarrays:
            time, iteration, temperature, value, best_value

        """

        # init
        base_value = self._problem.evaluate()
        self._problem.set_as_best(base_value)

        # init iteration (used to vount the amount of iterations)
        iteration = 0

        # add to data
        self._data_append(self.data, iteration, self._temperature, base_value,
                          self._problem.best_order_value)

        # init termination criterion
        self._termination_criterion.check_first_value(base_value)
        self._termination_criterion.start_timing()

        # main loop
        while self._termination_criterion.keep_running():

            # performs iterations at the current temperature
            #
            # it's a for loop that also terminates if the
            # termination criterion says it should.
            for i in range(
                    self._iterations_for_temp_f.get_iterations(
                        self._temperature)):

                if not self._termination_criterion.keep_running():
                    break

                # get and evaluate move
                move = self._problem.get_random_move()
                delta = self._problem.evaluate_move(move)

                # accept or reject move
                if self._is_improvement(0, delta):

                    # better than the current state --> accept
                    self._problem.move(move)
                    base_value = base_value + delta

                    # log if necessary
                    if delta is not 0:
                        self._log_improvement(base_value)

                    # check if best state
                    if self._is_better(
                            self._problem.best_order_value, base_value):
                        self._problem.set_as_best(base_value)

                    # let termination criterion check the new value
                    self._termination_criterion.check_new_value(base_value)

                    # add to data
                    self._data_append(self.data, iteration, self._temperature,
                                      base_value,
                                      self._problem.best_order_value)

                else:

                    # worse than current state --> use acceptance function.
                    if self._acceptance_function.accept(
                            delta, self._temperature):
                        self._problem.move(move)
                        base_value = base_value + delta

                        # log if necessary
                        self._log_passed_worse(base_value)

                        # let termination criterion check the new value
                        self._termination_criterion.check_new_value(base_value)

                        # add to data
                        self._data_append(self.data, iteration,
                                          self._temperature,
                                          base_value,
                                          self._problem.best_order_value)

                iteration += 1
                self._termination_criterion.iteration_done()

            # lowers the current temperature
            self._temperature = self._cooling_function.next_temperature(
                self._temperature)

            self._termination_criterion.check_variable(self._temperature)

        # last data point
        self._data_append(self.data, iteration, self._temperature,
                          base_value, self._problem.best_order_value)

        # if we have data:
        # convert data to something easier to plot
        if self.data is not None:

            # convert to tuple of list
            data = convert_data(self.data)

            # make namedtuple
            DataAsLists = namedtuple(
                'Data',
                ['time', 'iteration',
                 'temperature', 'value', 'best_value'])

            data = DataAsLists(data[0], data[1], data[2], data[3], data[4])

        else:
            data = None

        # return results

        Results = namedtuple('Results', ['best_order', 'best_value', 'data'])

        return Results(self._problem.best_order,
                       self._problem.best_order_value,
                       data)

    def reset(self):
        """Resets the object back to it's state after init.

        Raises
        ------
        NotImplementedError
            If the problem or termination criterion have no reset method
            implemented.

        Examples
        --------
        Minimising example, reset after run:

        .. doctest::

            >>> import numpy
            >>> import random
            >>> from lclpy.localsearch.move.tsp_array_swap import TspArraySwap
            >>> from lclpy.localsearch.simulatedannealing.simulated_annealing \\
            ...     import SimulatedAnnealing
            >>> from lclpy.localsearch.simulatedannealing.geometric_cooling_function \\
            ...     import GeometricCoolingFunction
            >>> from lclpy.localsearch.simulatedannealing.cnst_iterations_temp_function \\
            ...     import CnstIterationsTempFunction
            >>> from lclpy.evaluation.tsp_evaluation_function \\
            ...     import TspEvaluationFunction
            >>> from lclpy.termination.min_temperature_termination_criterion \\
            ...     import MinTemperatureTerminationCriterion
            >>> from lclpy.problem.array_problem import ArrayProblem
            ... # seed random
            ... # (used here to always get the same output, this obviously is not
            ... #                                  needed in your implementation.)
            >>> random.seed(0)
            ... # init problem
            >>> distance_matrix = numpy.array(
            ... [[0, 2, 5, 8],
            ...  [2, 0, 4, 1],
            ...  [5, 4, 0, 7],
            ...  [8, 1, 7, 0]])
            >>> size = distance_matrix.shape[0]
            >>> move = TspArraySwap(size)
            >>> evaluation = TspEvaluationFunction(distance_matrix, move)
            >>> problem = ArrayProblem(evaluation, move, size)
            ... # init termination criterion
            >>> termination_criterion = MinTemperatureTerminationCriterion()
            ... # init cooling function
            >>> cooling_func = GeometricCoolingFunction()
            ... # init CnstIterationsTempFunction
            ... # (determines amount of itertions in function of the temperature)
            >>> i_for_temp = CnstIterationsTempFunction()
            ... # init SimulatedAnnealing
            >>> algorithm = SimulatedAnnealing(problem, termination_criterion,
            ...                                cooling_func, i_for_temp, logging=False)
            ... # state before running
            >>> algorithm._start_temperature
            2000
            >>> algorithm._temperature
            2000
            >>> algorithm._problem._order
            array([0, 1, 2, 3])
            >>> algorithm._termination_criterion.keep_running()
            True
            >>> # run algorithm
            >>> algorithm.run()
            Results(best_order=array([0, 2, 3, 1]), best_value=15, data=None)
            >>> # tests reset
            >>> # before reset
            >>> algorithm._start_temperature
            2000
            >>> algorithm._temperature
            8.456565170490649
            >>> algorithm._problem._order
            array([0, 2, 1, 3])
            >>> algorithm._termination_criterion.keep_running()
            False
            >>> # reset
            >>> algorithm.reset()
            ... # after reset
            >>> algorithm._start_temperature
            2000
            >>> algorithm._temperature
            2000
            >>> algorithm._problem._order
            array([0, 1, 2, 3])
            >>> algorithm._termination_criterion.keep_running()
            True

        """

        self._temperature = self._start_temperature
        self._problem.reset()
        self._termination_criterion.reset()

        if self.data is not None:
            self.data = []



import numpy
from collections import namedtuple


def read_qaplib(filename):
    """Converts the qaplib file format to useable data structures.

    This function can be used to convert files in the qaplib format to useable
    data structures.

    Parameters
    ----------
    filename : str
        absolute or relative path to the file that contains the data.

    Returns
    -------
    distance_matrix : numpy.ndarray
        The distance matrix for the problem.
    flow_matrix : numpy.ndarray
        The flow matrix for the problem.
    dictionary : {int : int}
        A dictionary that can convert a position the distance matrix to the
        name given in the data. This dict can be useful when generating output.

    Examples
    --------
    Read "testfile.tsp" from the map "data" in the current working directory:

    .. code-block:: python

        from lclpy.io.qaplib import read_qaplib
        read_tsplib('data/testfile.tsp')

    """

    data = []
    # read the data
    with open(filename) as f:
        for line in f:
            splitted = line.split()

            if len(splitted) > 0:

                for i in splitted:
                    data.append(int(i))

    # create iterator

    size = data[0]

    # remove size from matrix
    data = data[1:]

    # create the 2 matrixes

    half = int(len(data) / 2)
    matrix_1 = numpy.array(data[:half], dtype=numpy.int_)
    matrix_1 = matrix_1.reshape(size, size)

    matrix_2 = numpy.array(data[half:], dtype=numpy.int_)
    matrix_2 = matrix_2.reshape(size, size)

    # check if matrix_1 has the least zeroes
    # if this is the case, we assume it's the distance matrix
    # if this isn't the case, we assume matrix_2 is the distance matrix
    if numpy.count_nonzero(matrix_1) < numpy.count_nonzero(matrix_2):
        matrix_1, matrix_2 = matrix_2, matrix_1

    # generate dictionary (we start counting from 0, not 1)
    dictionary = {}

    for i in range(size):
        dictionary[i] = i + 1

    TsplibData = namedtuple(
        'TsplibData', ['distance_matrix', 'flow_matrix', 'dictionary'])
    return TsplibData(matrix_1, matrix_2, dictionary)



import csv
import numpy
import collections
import os


def read_csv(filename, dtype="int"):
    """Converts the csv file format to useable data structures.

    Note that the csv file should only contain the cost matrixes.
    Different matrixes should be seperated with an empty line. A maximum of 2
    distance matrixes can be read.

    Parameters
    ----------
    filename : str
        Absolute or relative path to the file that contains the data.
    datatype : str, optional
        Determines the data type of the returned matrix(es).
        At the time, 2 datatypes are supported "float" and "int".
        The default float and int are respectively used if one of this
        parameter is used. The default is "int".

    Returns
    -------
    matrix_1 : numpy.ndarray
        A matrix for the problem.
    matrix_2 : numpy.ndarray
        A matrix for the problem, note that this matriw won't always be
        returned.

    Examples
    --------
    Read "testfile.csv" from the map "data" in the current directory:

    .. code-block:: python

        from lclpy.io.csv_io import read_csv
        read_csv('data/testfile.csv')

    """

    if dtype is "int":
        datatype = numpy.int_
    elif dtype is "float":
        datatype = numpy.float_

    matrix_1 = []
    matrix_2 = []

    current_matrix = matrix_1

    with open(filename, newline='') as file:

        spamreader = csv.reader(file, quoting=csv.QUOTE_NONNUMERIC)

        for row in spamreader:

            if len(row) == 0:
                current_matrix = matrix_2
                continue

            current_matrix.append(row)

    if len(matrix_2) == 0:
        CsvData = collections.namedtuple('CsvData', ['matrix_1'])
        return CsvData(numpy.array(matrix_1, dtype=datatype))

    else:
        CsvData = collections.namedtuple('CsvData', ['matrix_1', 'matrix_2'])
        return CsvData(numpy.array(matrix_1, dtype=datatype),
                       numpy.array(matrix_2, dtype=datatype))


def write_csv(result, filename):
    """Writes the result of an algorithm to a csv file.

    Parameters
    ----------
    result : collections.namedtuple
        The result from a localsearch algorithm
    filename : str
        Absolute or relative path to the file that one wishes to write too.
        It doesn't need to exist.

    """

    with open(filename, 'w', newline='') as file:
        writer = csv.writer(file)

        writer.writerow(("best state", ))
        writer.writerow(result[0])
        writer.writerow(tuple())

        writer.writerow(("best value", ))
        writer.writerow((result[1], ))
        writer.writerow(tuple())

        if result[2] is not None:

            writer.writerow(("data", ))
            data_size = len(result[2]._fields)
            size = len(result[2][0])
            writer.writerow(result[2]._fields)

            for i in range(size):

                data_row = []

                for j in range(data_size):
                    data_row.append(result[2][j][i])

                writer.writerow(data_row)


def write_benchmark_csv(benchmark_result, pathname,
                        algorithm_names=None, problem_names=None):
    """Writes the result of a benchmark to several csv files in a single map.

    Note that the name of the map can be chosen, but the csv files will always
    be named as follows: \n
    "alg\_" + algorithm_name + "_problem_" + problem_name + _run_ + run +
    ".csv" \n
    The runs always start from 0.

    Parameters
    ----------
    result : list of list of list of collections.namedtuple
        The result from a benchmark.
    pathname : str
        Absolute or relative path to the map that one wishes to write too.
        It doesn't need to exist.
    algorithm_names : list of str or tuple of str, optional
        Must be the same length as the algorithm list given to the original
        benchmark. The indices of a name corresponds with the indices of the
        algorithms given to the original benchmark.
    problem_names : list of str or tuple of str, optional
        Must be the same length as the problem list given to the original
        benchmark. The indices of a name corresponds with the indices of the
        problems given to the original benchmark.

    """

    if algorithm_names is None:
        algorithm_names = tuple(range(len(benchmark_result)))

    if problem_names is None:
        problem_names = tuple(range(len(benchmark_result[0])))

    os.makedirs(pathname, exist_ok=True)

    for i in range(len(benchmark_result)):
        for j in range(len(benchmark_result[0])):
            for k in range(len(benchmark_result[0][0])):

                filename = 'alg_' + str(algorithm_names[i]) + \
                    '_problem_' + \
                    str(problem_names[j]) + '_run_' + str(k) + '.csv'

                path = os.path.join(pathname, filename)

                write_csv(benchmark_result[i][j][k], path)



from lclpy.termination.abstract_termination_criterion \
    import AbstractTerminationCriterion
from lclpy.aidfunc.is_improvement_func import bigger, smaller


class MustImproveTerminationCriterion(AbstractTerminationCriterion):
    """Criterion to terminate after an iteration without improvement.

    Parameters
    ----------
    minimise : bool, optional
        If the goal is to minimise the evaluation function, this should be
        True. If the goal is to maximise the evaluation function, this should
        be False. The default is True.

    Attributes
    ----------
    _old_best_value: int
        The last value. Is initialised as infinite (minimise = True)
        or minus infinite (minimise = False)
    _function : function
        The function used to judge if a value is an improvement.
    _run : bool
        True if no worse value has been encountered, False if this isn't the
        case.

    Examples
    --------
    Smaller values are considered improvements(default). 8 iterations with
    improvement. After that there are no more improvements. The dataset
    eval_values is hardcoded:

    .. doctest::

        >>> import numpy
        >>> from lclpy.termination.must_improve_termination_criterion \\
        ...     import MustImproveTerminationCriterion
        ... # creation of an array that contains the values that will be given
        ... # to our termination criterion
        >>> eval_values = numpy.array([10, 9, 8, 7, 6, 5, 4, 3, 4, 3, 5, 12])
        ... # index is used to get values from the array.
        ... # index is also used to count the amount of iterations
        >>> index = 0
        ... # init
        >>> test = MustImproveTerminationCriterion()
        ... # loop
        >>> while test.keep_running():
        ...     pass # other code to execute
        ...     # check the next value
        ...     # not_used is only used to suppress the output of
        ...     # check_new_value
        ...     test.check_new_value(eval_values[index])
        ...     pass # other code to execute
        ...     # counting iterations + increment index
        ...     index += 1
        >>> index # == amount of iterations.
        9

    Bigger values are considered improvements. 5 iterations with
    improvement. After that there are no more improvements. The dataset
    eval_values is hardcoded:

    .. doctest::

        >>> import numpy
        >>> from lclpy.termination.must_improve_termination_criterion \\
        ...     import MustImproveTerminationCriterion
        ... # creation of an array that contains the values that will be given
        ... # to our termination criterion
        >>> eval_values = numpy.array([0, 2, 14, 15, 20, 3, 3, 4, 2, 1, 0, 12])
        ... # index is used to get values from the array.
        ... # index is also used to count the amount of iterations
        >>> index = 0
        ... # init
        >>> test = MustImproveTerminationCriterion(False)
        ... # loop
        >>> while test.keep_running():
        ...     pass # other code to execute
        ...     # check the next value
        ...     # not_used is only used to suppress the output of
        ...     # check_new_value
        ...     test.check_new_value(eval_values[index])
        ...     pass # more code to execute
        ...     # counting iterations + increment index
        ...     index += 1
        >>> index # == amount of iterations.
        6



    """

    def __init__(self, minimise=True):
        super().__init__()

        # init
        self._run = True

        # choose intial _old_best_value value + pick judge function
        if minimise:
            self._function = smaller
            self._old_best_value = float("inf")
        else:
            self._function = bigger
            self._old_best_value = float("-inf")

    def keep_running(self):
        """function to determine if the algorithm needs to continue running

        Returns
        -------
        bool
            The function returns True if there hasn't been encountered a
            non-improving value. The function returns False if a non-improving
            value is encountered.
        """
        return self._run

    def check_new_value(self, value):
        """function to be called after every calculation of the evaluation function.

        Parameters
        ----------
        value : int or long or float
            Is the new evaluation value of a solution.

        """

        if self._function(self._old_best_value, value):
            self._old_best_value = value
        else:
            self._run = False

    def reset(self):
        """Resets the object back to it's state after init.

        Examples
        --------
        Smaller values are considered improvements(default). 8 iterations with
        improvement. After that there are no more improvements. The dataset
        eval_values is hardcoded. After this, the criterion is resetted and the
        loop is repeated:

        .. doctest::

            >>> import numpy
            >>> from lclpy.termination.must_improve_termination_criterion \\
            ...     import MustImproveTerminationCriterion
            ... # creation of an array that contains the values that will be given
            ... # to our termination criterion
            >>> eval_values = numpy.array([10, 9, 8, 7, 6, 5, 4, 3, 4, 3, 5, 12])
            ... # init
            >>> test = MustImproveTerminationCriterion()
            ... # run 1
            ... #
            ... # index is used to get values from the array.
            ... # index is also used to count the amount of iterations
            >>> index = 0
            ... # loop
            >>> while test.keep_running():
            ...     pass # other code to execute
            ...     # check the next value
            ...     # not_used is only used to suppress the output of
            ...     # check_new_value
            ...     test.check_new_value(eval_values[index])
            ...     pass # other code to execute
            ...     # counting iterations + increment index
            ...     index += 1
            >>> index # == amount of iterations.
            9
            >>> # reset
            >>> test.reset()
            ... # run 2
            >>> index = 0
            ... # loop
            >>> while test.keep_running():
            ...     pass # other code to execute
            ...     # check the next value
            ...     # not_used is only used to suppress the output of
            ...     # check_new_value
            ...     test.check_new_value(eval_values[index])
            ...     pass # other code to execute
            ...     # counting iterations + increment index
            ...     index += 1
            >>> index # == amount of iterations.
            9



        """

        self._run = True

        # restore old_best value
        if self._function is smaller:
            # if minimising
            self._old_best_value = float("inf")
        else:
            # if maximising
            self._old_best_value = float("-inf")



from lclpy.evaluation.abstract_evaluation_function \
    import AbstractEvaluationFunction
from lclpy.evaluation.deltaeval.delta_eval_func import delta_eval_func


class TspEvaluationFunction(AbstractEvaluationFunction):
    """A class to evaluate TSP problems.

    This class contains the methods to evaluate the quality of a solution for a
    travelling salesman problem (TSP). The distance matrix is allowed to be
    asymmetric.

    Parameters
    ----------
    distance_matrix : numpy.ndarray
        The distance matrix of the tsp-problem. The weight from A to B does
        not need to be equal to the weight from B to A.
    move_function : AbstractMove, optional
        Only needs to be passed if one wishes to use delta evaluation.

    Attributes
    ----------
    _distance_matrix : numpy.ndarray
        The distance matrix of the tsp-problem.
    _size : int
        the amount of points to visit, is derived from the distance matrix.

    Examples
    --------
    A simple example:

    .. doctest::

        >>> import numpy
        >>> from lclpy.evaluation.tsp_evaluation_function \\
        ...     import TspEvaluationFunction
        ... # init
        >>> dist_matrix = numpy.array(
        ... [[0, 2, 9, 5],
        ...  [2, 0, 4, 6],
        ...  [9, 4, 0, 3],
        ...  [5, 6, 3, 0]])
        >>> eval_func = TspEvaluationFunction(dist_matrix)
        ... # tests
        >>> order = numpy.array([0, 1, 2, 3])
        >>> eval_func.evaluate(order)
        14
        >>> order = numpy.array([0, 3, 1, 2])
        >>> eval_func.evaluate(order)
        24
        >>> order = numpy.array([2, 0, 1, 3])
        >>> eval_func.evaluate(order)
        20


    """

    def __init__(self, distance_matrix, move_function=None):
        super().__init__()
        self._distance_matrix = distance_matrix
        self._size = distance_matrix.shape[0]

        if move_function is not None:
            self._delta_evaluate_object = delta_eval_func(self, move_function)
            self.delta_evaluate = self._delta_evaluate_object.delta_evaluate

    def get_problem_type(self):
        """Returns the problem type.

        Returns
        -------
        str
            The problem type.

        """

        return 'TSP'

    def evaluate(self, order):
        """Calculates an evaluation value for the function.

        Parameters
        ----------
        order : numpy.ndarray
            A 1 dimensional array that contains the order of the points to
            visit. All values are unique and are within the interval [0,size[.

        Returns
        -------
        int, float
            An indication of the quality of the solution, the lower this
            value, the better the quality.
        """

        # init value
        value = 0

        # add all distances to value
        for index in range(self._size - 1):
            value += self._distance_matrix[order[index]][order[index + 1]]

        value += self._distance_matrix[order[-1]][order[0]]

        return value

    def delta_evaluate(self, current_order, move):
        """Calculates the difference in quality if the move would be performed.

        Note that a move function needs to be passed to the constructor of
        evaluation function for the delta_evaluate to work. The move
        function also needs to have changed_distances and
        transform_next_index_to_current_index properly implemented:

        Parameters
        ----------
        current_order : numpy.ndarray
            A 1 dimensional array that contains the order of the points to
            visit. All values are unique and are within the interval [0,size[.
            This is the current order.
        move : tuple of int
            Contains the move one wishes to know the effects on the quality of.

        Returns
        -------
        int or float
            The difference in quality if the move would be performed.

        Raises
        ------
        NotImplementedError
            If no move_function was given in the constructor.

        Examples
        --------
        A simple example to demonstrate the working of delta_evaluate.
        Note that other move types than array_swap can be used:

        .. doctest::

            >>> import numpy
            >>> from lclpy.evaluation.tsp_evaluation_function \\
            ...     import TspEvaluationFunction
            >>> from lclpy.localsearch.move.tsp_array_swap \\
            ...     import TspArraySwap
            ... # init distance matrix
            >>> dist_matrix = numpy.array(
            ... [[0, 2, 9, 5],
            ...  [2, 0, 4, 6],
            ...  [9, 4, 0, 3],
            ...  [5, 6, 3, 0]])
            ... # init move function
            >>> move_func = TspArraySwap(4)
            ... # init evaluation function
            >>> eval_func = TspEvaluationFunction(dist_matrix, move_func)
            ... # init of the order
            >>> order = numpy.array([2, 0, 1, 3])
            ... # tests
            >>> eval_func.delta_evaluate(order, (1, 2))
            -6
            >>> eval_func.delta_evaluate(order, (1, 3))
            0
            >>> eval_func.delta_evaluate(order, (2, 3))
            4

        A more elaborate example:

        .. doctest::

            >>> import numpy
            >>> from lclpy.evaluation.tsp_evaluation_function \\
            ...     import TspEvaluationFunction
            >>> from lclpy.localsearch.move.array_swap \\
            ...     import ArraySwap
            ... # init distance matrix
            >>> dist_matrix = numpy.array(
            ... [[0, 2, 9, 5, 8, 1],
            ...  [2, 0, 4, 6, 1, 2],
            ...  [9, 4, 0, 3, 6, 3],
            ...  [5, 6, 3, 0, 7, 4],
            ...  [8, 1, 6, 7, 0, 5],
            ...  [1, 2, 3, 4, 5, 0]])
            ... # init move function
            >>> move_func = TspArraySwap(6)
            ... # init evaluation function
            >>> eval_func = TspEvaluationFunction(dist_matrix, move_func)
            ... # init of the order
            >>> order = numpy.array([0, 5, 2, 1, 3, 4])
            ... # tests
            >>> eval_func.delta_evaluate(order, (1, 4))
            -2
            >>> eval_func.delta_evaluate(order, (0, 1))
            3
            >>> eval_func.delta_evaluate(order, (2, 4))
            0


        """

        raise NotImplementedError



from lclpy.evaluation.deltaeval.delta_multi_neighbourhood \
    import delta_multi_neighbourhood
from lclpy.evaluation.deltaeval.delta_tsp import delta_tsp
from lclpy.evaluation.deltaeval.delta_qap import delta_qap


def delta_eval_func(problem_eval_func, move_func):
    """A function to retrieve classes and functions needed for delta evaluation.

    Parameters
    ----------
    problem_eval_func : AbstractEvaluationFunction
        The used evaluation function object.
    move_type : AbstractMove
        The used move object.

    Returns
    -------
    object
        A class with a function delta_evaluate to perform delta evaluation.

    """

    problem_type = problem_eval_func.get_problem_type()
    move_type = move_func.get_move_type()

    if move_type is 'multi_neighbourhood':
        return delta_multi_neighbourhood(problem_eval_func, move_func)
    if problem_type is 'TSP':
        return delta_tsp(problem_eval_func, move_func)
    elif problem_type is 'QAP':
        return delta_qap(problem_eval_func, move_func)
    else:
        raise NotImplementedError



from lclpy.aidfunc.error_func import _not_implemented


# base class for the tsp delta evaluation

class QAPDeltaEvaluate():
    """Class to perform delta-evaluation for QAP problems.

    Parameters
    ----------
    eval_func : AbstractEvaluationFunction
        The used evaluation function
    changed_points
        This function returns the pairs who would have an altered evaluation
        value due to the move.
    next_to_current
        This function transforms the indices so that they can be used as
        indices in the unaltered array, yet return the value they would have
        had if the move was actually performed and they were used as indices.

    Attributes
    ----------
    eval_func : AbstractEvaluationFunction
        The used evaluation function
    changed_locations
        This function returns the locations who would have an altered
        evaluation value due to the move.
    next_to_current
        This function transforms the indices so that they can be used as
        indices in the unaltered array, yet return the value they would have
        had if the move was actually performed and they were used as indices.

    """

    def __init__(self, eval_func, changed_locations, next_to_current):
        self.eval_func = eval_func
        self.changed_locations = changed_locations
        self.next_to_current = next_to_current

    def delta_evaluate(self, current_order, move):
        """Calculates the difference in quality if the move would be performed.

        Parameters
        ----------
        eval_func : AbstractEvaluationFunction
            The object the delta-evaluation is calculated for.
        current_order : numpy.ndarray
            A 1 dimensional array that contains the order of the points to
            visit. All values are unique and are within the interval [0,size[.
            This is the current order.
        move : tuple of int
            Contains the move one wishes to know the effects on the quality of.

        Returns
        -------
        int or float
            The difference in quality if the move would be performed.

        """

        # get the changed locations
        # these are represented as a set of ints
        changed = self.changed_locations(move)

        # init values
        next_solution_value = 0
        current_solution_value = 0

        visited = set()

        # calculate the value for all changed locations
        for location in changed:

            visited.add(location)

            next_location = self.next_to_current(location, move)

            for i in range(self.eval_func._size):

                if i not in visited:

                    current_solution_value += \
                        self.eval_func._distance_matrix[location][i] * \
                        self.eval_func._flow_matrix[
                            current_order[location]][current_order[i]]

                    next_i = self.next_to_current(i, move)

                    next_solution_value += \
                        self.eval_func._distance_matrix[location][i] * \
                        self.eval_func._flow_matrix[
                            current_order[next_location]][current_order[next_i]]

        return next_solution_value - current_solution_value


# functions for array_swap

def array_swap_changed_locations(move):
    """Aid function for delta evaluation.

    Works with the array_swap move type.
    This function returns the locations who would have an altered evaluation
    value due to the move.

    Parameters
    ----------
    move : tuple of int
        A tuple of 2 ints that represents a single unique move.

    Returns
    -------
    tuple of int
        A tuple containing the locations that would have an altered evaluation
        value due to the move.

    Examples
    --------
    Some simple examples to demonstrate the behaviour:

    .. doctest::

        >>> from lclpy.evaluation.deltaeval.delta_qap \\
        ...     import array_swap_changed_locations \\
        ...         as changed_locations
        ... # tests
        >>> changed_locations((4, 8))
        (4, 8)
        >>> changed_locations((4, 9))
        (4, 9)
        >>> changed_locations((0, 8))
        (0, 8)
        >>> changed_locations((0, 9))
        (0, 9)

    """

    return move


def array_swap_transform_next_index_to_current_index(position, move):
    """Transforms the position depending on the move.

    Works with the array_swap move type.
    This function transforms the position so that it can be used as the indice
    in the unaltered array, yet return the value it would have had if the move
    was actually performed and the position was used as indice.

    Parameters
    ----------
    position : int
        The index that one wants to use in the array if the move was performed.
    move : tuple of int
        A tuple with that represents a single, unique move.

    Returns
    -------
    int
        The index in the unaltered array that has the same value as the
        location in an array where the move was performed.

    Examples
    --------
    Some simple examples, the move remains the same, but the position changes:

    .. doctest::

        >>> from lclpy.evaluation.deltaeval.delta_qap \\
        ...     import array_swap_transform_next_index_to_current_index \\
        ...         as transform_next_index_to_current_index
        ... # tests
        >>> transform_next_index_to_current_index(0, (1, 3))
        0
        >>> transform_next_index_to_current_index(1, (1, 3))
        3
        >>> transform_next_index_to_current_index(2, (1, 3))
        2
        >>> transform_next_index_to_current_index(3, (1, 3))
        1
        >>> transform_next_index_to_current_index(4, (1, 3))
        4

    """

    # transform frm so it returns the value that from would have if the
    # move was performed.
    if position == move[0]:
        position = move[1]
    elif position == move[1]:
        position = move[0]

    return position


# functions for array_reverse_order

def array_reverse_order_changed_locations(move):
    """Aid function for delta evaluation.

    Works with the array_swap move type.
    This function returns the locations who would have an altered evaluation
    value due to the move.

    Parameters
    ----------
    move : tuple of int
        A tuple of 2 ints that represents a single unique move.

    Returns
    -------
    range
        A range containing the locations that would have an altered evaluation
        value due to the move.

    Examples
    --------
    Some simple examples to demonstrate the behaviour:

    .. doctest::

        >>> from lclpy.evaluation.deltaeval.delta_qap \\
        ...     import array_reverse_order_changed_locations \\
        ...         as changed_locations
        ... # tests
        >>> changed_locations((4, 5))
        range(4, 6)
        >>> changed_locations((4, 9))
        range(4, 10)
        >>> changed_locations((0, 8))
        range(0, 9)

    """

    return range(move[0], move[1] + 1)


def array_reverse_order_transform_next_index_to_current_index(position, move):
    """Transforms the position depending on the move.

    Works with the array_swap move type.
    This function transforms the position so that it can be used as the indice
    in the unaltered array, yet return the value it would have had if the move
    was actually performed and the position was used as indice.

    Parameters
    ----------
    position : int
        The index that one wants to use in the array if the move was performed.
    move : tuple of int
        A tuple with that represents a single, unique move.

    Returns
    -------
    int
        The index in the unaltered array that has the same value as the
        location in an array where the move was performed.

    Examples
    --------
    Some simple examples, the move remains the same, but the position changes:

    .. doctest::

        >>> from lclpy.evaluation.deltaeval.delta_qap \\
        ...     import array_reverse_order_transform_next_index_to_current_index \\
        ...         as transform_next_index_to_current_index
        ... # tests
        >>> transform_next_index_to_current_index(0, (1, 4))
        0
        >>> transform_next_index_to_current_index(1, (1, 4))
        4
        >>> transform_next_index_to_current_index(2, (1, 4))
        3
        >>> transform_next_index_to_current_index(3, (1, 4))
        2
        >>> transform_next_index_to_current_index(4, (1, 4))
        1
        >>> transform_next_index_to_current_index(5, (1, 4))
        5

    """

    # check if position is altered by the move

    if (position >= move[0]) & (position <= move[1]):

        # alter the position
        offset = position - move[0]
        position = move[1] - offset

    return position


# shared delta_evaluate function for TSP problems

def delta_evaluate(eval_func, current_order, move):
    """Calculates the difference in quality if the move would be performed.

    Parameters
    ----------
    eval_func : AbstractEvaluationFunction
        The object the delta-evaluation is calculated for.
    current_order : numpy.ndarray
        A 1 dimensional array that contains the order of the points to
        visit. All values are unique and are within the interval [0,size[.
        This is the current order.
    move : tuple of int
        Contains the move one wishes to know the effects on the quality of.

    Returns
    -------
    int or float
        The difference in quality if the move would be performed.

    """

    # get the changed locations
    # these are represented as a set of ints
    changed = eval_func._changed_distances(move)

    # init values
    next_solution_value = 0
    current_solution_value = 0

    visited = set()

    # calculate the value for all changed locations
    for location in changed:

        visited.add(location)

        next_location = \
            eval_func._transform_next_index_to_current_index(location, move)

        for i in range(eval_func._size):

            if i not in visited:

                current_solution_value += \
                    eval_func._distance_matrix[location][i] * \
                    eval_func._flow_matrix[
                        current_order[location]][current_order[i]]

                next_i = eval_func._transform_next_index_to_current_index(
                    i, move)

                next_solution_value += \
                    eval_func._distance_matrix[location][i] * \
                    eval_func._flow_matrix[
                        current_order[next_location]][current_order[next_i]]

    return next_solution_value - current_solution_value


# The method to return the other stuff

def delta_qap(eval_func, move_func):
    """Returns delta-eval class for a QAP problem.

    Note that if a method for the problem can't be found, that a placeholder
    method will be used. This method will raise a NotImplementedError when
    called.

    Parameters
    ----------
    eval_func : AbstractEvaluationFunction
        The used evaluation function object.
    move_func : AbstractMove
        The used move object.

    Returns
    -------
    QAPDeltaEvaluate
        Class useable for delta evaluation of TSP problems.

    """

    move_type = move_func.get_move_type()

    if move_type is 'array_swap':
        return QAPDeltaEvaluate(eval_func,
                                array_swap_changed_locations,
                                array_swap_transform_next_index_to_current_index)
    if move_type is 'array_reverse_order':
        return QAPDeltaEvaluate(eval_func,
                                array_reverse_order_changed_locations,
                                array_reverse_order_transform_next_index_to_current_index)
    else:
        raise NotImplementedError



from lclpy.localsearch.acceptance.abstract_acceptance_function \
    import AbstractAcceptanceFunction
import random
import math


class SimulatedAnnealingAcceptanceFunction(AbstractAcceptanceFunction):
    """An acceptance function for simulated annealing.

    A smaller quality value is assumed to be better than a bigger one.

    Parameters
    ----------
    diff_multiplier : int or float, optional
        The delta_value will be multiplied by this multiplier. A bigger value
        leads to more solutions being accepted, while a smaller value will
        lead to less solutions being accepted.
        Positive values will initialise the class for minising, negative values
        will intialise the function for maximising.
        The default value is 1.
    multiplier : int or float, optional
        Will be multiplied with the whole probability. Must be positive.
        Should be in the interval ]0,1]. This multiplier can be used to
        decrease the odds of values being accepted. While it is possible to use
        a multiplier greater than 1, this might cause weird behaviour.
        The default value is 1.

    Attributes
    ----------
    _diff_multiplier : int or float
        The delta_value will be multiplied by this multiplier.
    _multiplier : int or float
        Is multiplied with the whole probability.

    Examples
    --------
    A simple example of the default behaviour, delta (200) and
    temperature (1000) are arbitrarily chosen, they have little meaning in
    this example:

    .. doctest::

        >>> import random
        >>> from lclpy.localsearch.acceptance.simulated_annealing_acceptance_function \\
        ...     import SimulatedAnnealingAcceptanceFunction
        ... # set seed of random
        ... # this isn't needed, it's only used to make sure the results will
        ... # always be the same.
        >>> random.seed(2)
        ... # init
        >>> test = SimulatedAnnealingAcceptanceFunction()
        ... # tests
        >>> test.accept(200, 1000)
        False
        >>> test.accept(200, 1000)
        False
        >>> test.accept(200, 1000)
        True
        >>> test.accept(200, 1000)
        True
        >>> test.accept(200, 1000)
        False

    """

    def __init__(self, diff_multiplier=1, multiplier=1):
        super().__init__()

        self._diff_multiplier = diff_multiplier
        self._multiplier = multiplier

    def accept(self, delta_value, temperature):
        """Function to reject or accept certain potential solutions.

        In most cases, a value better than the current value will always be
        accepted. When this is the case, it might be wise to use an
        if-statement to handle those instead of passing them to this method.

        The chance of a state being accepted is equal to:

        .. math::

            \\_multiplier*e^{\\dfrac{\\_diff\\_multiplier*delta\\_value}{temperature}}

        .. The formula:
           _multiplier*e^(-(_diff_multiplier*delta_value)/temperature)

        Parameters
        ----------
        delta_value : int or float
            The difference in quality between 2 possible solutions.
        temperature : int or float
            A value that determines the chance of a state being accepted.
            How bigger this value, how higher the chance of a solution being
            accepted.

        Returns
        -------
        bool
            True if the solution is accepted, False if the solution isn't
            accepted.

        """

        # calculate probability
        probability = self._multiplier * \
            math.exp(-(self._diff_multiplier * delta_value) / temperature)

        # generates a random number in the interval [0, 1[
        random_number = random.random()

        return probability > random_number



from time import perf_counter


def add_to_data_func(data, *args):
    """A function to add values to a list with a timestamp.

    The data will be kept into a tuple and added to the list.
    Note that a timestamp is always added as the first item in the tuple. This
    timestamp, however has no fixed epoch. So only differences in time have a
    real meaning.

    Parameters
    ----------
    data : list
        The list one wishes to append the data to.
    *args
        The data one wishes to append to data.

    """

    data.append((perf_counter(),) + args)



from lclpy.evaluation.abstract_evaluation_function \
    import AbstractEvaluationFunction
from lclpy.evaluation.deltaeval.delta_eval_func import delta_eval_func


class QuadraticAssignmentEvaluationFunction(AbstractEvaluationFunction):
    """A class to evaluate QAP problems.

    This class contains the methods to evaluate the quality of a solution for a
    quadratic assignment problem (QAP). In the implementation, it was assumed
    that both the distance and the flow matrix are symmetric.

    Parameters
    ----------
    distance_matrix : numpy.ndarray
        The distance matrix of the problem. Should be symmetric.
    flow_matrix : numpy.ndarray
        The flow matrix for the problem. Should be symmetric.
    move_function : AbstractMove, optional
        Only needs to be passed if one wishes to use delta evaluation.

    Attributes
    ----------
    _distance_matrix : numpy.ndarray
        The distance matrix of the problem.
    _flow_matrix : numpy.ndarray
        The flow matrix of the problem.
    _size : int
        The amount of locations, derived from the distance matrix.

    Examples
    --------
    A simple example:

    .. doctest::

        >>> import numpy
        >>> from lclpy.evaluation.quadratic_assignment_evaluation_function \\
        ...     import QuadraticAssignmentEvaluationFunction
        ... # init distance matrix
        >>> dist_matrix = numpy.array(
        ... [[0, 2, 9, 5],
        ...  [2, 0, 4, 6],
        ...  [9, 4, 0, 3],
        ...  [5, 6, 3, 0]])
        ... # init flow matrix
        >>> flow_matrix = numpy.array(
        ... [[0, 2, 0, 0],
        ...  [2, 0, 4, 0],
        ...  [0, 4, 0, 8],
        ...  [0, 0, 8, 0]])
        ... # init evaluation function
        >>> eval_func = QuadraticAssignmentEvaluationFunction(dist_matrix,
        ...                                                   flow_matrix)
        ... # tests
        >>> order = numpy.array([0, 1, 2, 3])
        >>> eval_func.evaluate(order)
        44
        >>> order = numpy.array([0, 3, 1, 2])
        >>> eval_func.evaluate(order)
        78
        >>> order = numpy.array([2, 1, 0, 3])
        >>> eval_func.evaluate(order)
        56

    """

    def __init__(self, distance_matrix, flow_matrix, move_function=None):
        super().__init__()

        self._size = distance_matrix.shape[0]
        self._distance_matrix = distance_matrix
        self._flow_matrix = flow_matrix

        if move_function is not None:
            self._delta_evaluate_object = delta_eval_func(self, move_function)
            self.delta_evaluate = self._delta_evaluate_object.delta_evaluate

    def get_problem_type(self):
        """Returns the problem type.

        Returns
        -------
        str
            The problem type.

        """

        return 'QAP'

    def evaluate(self, order):
        """Calculates an evaluation value for the function.

        Parameters
        ----------
        order : numpy.ndarray
            A 1 dimensional array that maps facilities on locations. The
            index respresents a location, the corresponding value represents
            a facility.

        Returns
        -------
        int or float
            an indication of the quality of current_data

        """

        value = 0

        # all distances need to be checked once
        for i in range(self._size):
            for j in range(i + 1, self._size):
                value += self._distance_matrix[i][j] * \
                    self._flow_matrix[order[i]][order[j]]

        return value

    def delta_evaluate(self, current_order, move):
        """Evaluates the difference in quality between two solutions.

        The two compared solutions are the current solution and the solution
        if the move was performed. The move is not actually performed.

        This function does not need to be implemented. One should only
        consider to implement and use it if a delta evaluation is faster than
        the regular evaluate function or if it needs to be implemented to work
        with existing code.

        Parameters
        ----------
        current_order : numpy.ndarray
            A 1 dimensional array that maps facilities on locations. The
            index respresents a location, the corresponding value represents
            a facility.

        move : tuple of int
            Contains the move one wishes to know the effects on the quality of.

        Returns
        -------
        int or float
            The difference in quality if the move would be performed.

        Raises
        ------
        NotImplementedError
            If no move_function was given in the constructor.

        Examples
        --------
        A simple example to demonstrate the working of delta_evaluate.
        Note that other move types than array_swap can be used:

        .. doctest::

            >>> import numpy
            >>> from lclpy.evaluation.quadratic_assignment_evaluation_function \\
            ...     import QuadraticAssignmentEvaluationFunction
            >>> from lclpy.localsearch.move.array_swap \\
            ...     import ArraySwap
            ... # init distance matrix
            >>> dist_matrix = numpy.array(
            ... [[ 0, 22, 53, 53],
            ...  [22,  0, 40, 62],
            ...  [53, 40,  0, 55],
            ...  [53, 62, 55,  0]])
            ... # init flow matrix
            >>> flow_matrix = numpy.array(
            ... [[0, 3, 0, 2],
            ...  [3, 0, 0, 1],
            ...  [0, 0, 0, 4],
            ...  [2, 1, 4, 0]])
            ... # init move function
            >>> move_func = ArraySwap(4)
            ... # init evaluation function
            >>> eval_func = QuadraticAssignmentEvaluationFunction(
            ...     dist_matrix, flow_matrix, move_func)
            ... # tests
            >>> order = numpy.array([0, 1, 2, 3])
            >>> eval_func.delta_evaluate(order, (0, 1))
            9
            >>> eval_func.delta_evaluate(order, (0, 2))
            50
            >>> eval_func.delta_evaluate(order, (0, 3))
            72
            >>> eval_func.delta_evaluate(order, (1, 2))
            114
            >>> eval_func.delta_evaluate(order, (2, 3))
            -22

        """

        raise NotImplementedError



"""This module contains the classes and functions for delta-evaluation with a
MultiNeighbourhood.

"""


# class for multi neighbourhood delta evaluation

class MultiMoveDeltaEvaluate():
    """Class to perform delta evaluation with a MultiNeighbourhood.

    Parameters
    ----------
    problem_eval_func : AbstractEvaluationFunction
        The used evaluation function object.
    move_type : MultiNeighbourhood
        The used move object.

    Attributes
    ----------
    delta_eval_classes : tuple or list
        Contains the classes used to perform delta evaluation for each
        move type.
    delta_evaluation_func : tuple or list
        Contains the functions of the classes in delta_eval_classes that are
        called when delta-evaluation is performed.

    """

    def __init__(self, eval_func, move_func):
        # import is here because of circular dependency.
        from lclpy.evaluation.deltaeval.delta_eval_func \
            import delta_eval_func

        self.delta_eval_classes = []

        for move in move_func._move_func_list:
            self.delta_eval_classes.append(delta_eval_func(eval_func, move))

        self.delta_eval_classes = tuple(self.delta_eval_classes)

        self.delta_evaluation_func = []

        for instance in self.delta_eval_classes:
            self.delta_evaluation_func.append(instance.delta_evaluate)

        self.delta_evaluation_func = tuple(self.delta_evaluation_func)

    def delta_evaluate(self, current_order, move):

        return self.delta_evaluation_func[move[0]](current_order, move[1])


# The method to return the class

def delta_multi_neighbourhood(eval_func, move_func):
    """Returns delta-eval class for a problem with a multi-neighbourhood.

    Note that if no methods for a move or the evaluation function can be found,
    that a placeholder method will be used. The placeholder methods will raise
    a NotImplementedError when called.

    Parameters
    ----------
    eval_func : AbstractEvaluationFunction
        The used evaluation function.
    move_func : AbstractMove
        The used move function.

    Returns
    -------
    MultiMoveDeltaEvaluate
        Class useable for delta evaluation of TSP problems.

    """

    return MultiMoveDeltaEvaluate(eval_func, move_func)



from collections import deque


class TabuList(object):
    """Implements a tabu list for use with TabuSearch.

    When using big items, it would be wise to save space by hashing them.

    Parameters
    ----------
    length : int
        The maximal amount of items in the tabu list.

    Attributes
    ----------
    _list : deque
        A list that contains hashes of all the items that are considered part
        of the tabu list.

    Examples
    --------
    A simple example, the lists are converted to tuples, because they're
    mutable objects and thus can't be hashed with the default implementation of
    hash:

    .. doctest::

        >>> from lclpy.localsearch.tabusearch.tabu_list import TabuList
        >>> test = TabuList(3)
        >>> test.add(tuple([0, 1, 2]))
        >>> test.add(tuple([1, 0, 2]))
        >>> test.add(tuple([0, 2, 1]))
        >>> test.add(tuple([2, 1, 0]))
        >>> test.contains(tuple([0, 1, 2]))
        False
        >>> test.contains(tuple([1, 0, 2]))
        True
        >>> test.contains(tuple([0, 2, 1]))
        True
        >>> test.contains(tuple([2, 1, 0]))
        True

    """

    def __init__(self, length):
        super().__init__()
        self._list = deque(maxlen=length)

    def add(self, item):
        """Adds an item to the tabu list.

        Parameters
        ----------
        item
            A hashable, immutable object.

        """

        self._list.append(item)

    def contains(self, item):
        """A method that checks if an item is in the tabu list.

        Parameters
        ----------
        item
            A hashable, immutable object.

        Returns
        -------
        bool
            Will return true if the item is in the tabu list, false if it
            isn't in the tabu list.

        """
        return item in self._list



from abc import ABC, abstractmethod


class AbstractLocalSearchProblem(ABC):
    """Template to create Problem-classes for localsearch implemenatations.

    This Class is meant to be used as a template.
    It is supposed to be used when constructing your own solution objects for a
    specific problem.

    """

    def __init__(self):
        super().__init__()

    @abstractmethod
    def move(self, move):
        """Performs a move.

        Parameters
        ----------
        move : tuple of int
            A representation of a move.

        """

        pass

    @abstractmethod
    def undo_move(self, move):
        """Undoes a move.

        Parameters
        ----------
        move : tuple of int
            A representation of the move one wishes to undo.

        """

        pass

    @abstractmethod
    def get_moves(self):
        """An iterable that returns all valid moves in the complete neighbourhood.

        Yields
        ------
        move : tuple of int
            A representation of the next move in the neighbourhood.

        """

        pass

    @abstractmethod
    def get_random_move(self):
        """A function to return a random move from the complete neighbourhood.

        Returns
        -------
        tuple of int
            A representation of the move.

        """

        pass

    def evaluate_move(self, move):
        """Evaluates the quality gained or lost by a potential move.

        This function does not need to be implemented.

        Parameters
        ----------
        move : tuple of int
            Represents a unique valid move.

        Returns
        -------
        int or float
            The change in value of the eval-function if the move is performed.

        """

        raise NotImplementedError

    @abstractmethod
    def evaluate(self):
        """Evaluates the current state of the solution.

        Returns
        -------
        int or float
            An indication of the quality of the current state.

        """

        pass

    @abstractmethod
    def set_as_best(self):
        """Saves the current state as the best found state."""

        pass

    @abstractmethod
    def state(self):
        """Returns an immutable hashable object that describes the current state.

        Returns
        -------
        tuple
            A representation of the current state.

        """

        pass

    def first_neighbourhood(self):
        """Changes the current neighbourhood to the first neighbourhood.

        Note that this function will only be useable if the neighbourhood given
        to the constructor is a MultiNeighbourhood.

        Raises
        ------
        WrongMoveTypeError
            If the move_function isn't a MultiNeighbourhood.

        """

        raise NotImplementedError

    def next_neighbourhood(self):
        """Changes the current neighbourhood to the next neighbourhood.

        Note that this function will only be useable if the neighbourhood given
        to the constructor is a MultiNeighbourhood.
        If this function is called when the last neighbourhood is the current
        neighbourhood, the last neighbourhood will remain the current
        neighbourhood and an exception will be raised.

        Raises
        ------
        NoNextNeighbourhood
            If there is no next neighbourhood. This is simply an indication
            that the current neighbourhood was the last neighbourhood.
        WrongMoveTypeError
            If the move_function isn't a MultiNeighbourhood.

        """

        raise NotImplementedError

    def previous_neighbourhood(self):
        """Changes the current neighbourhood to the previous neighbourhood.

        Note that this function will only be useable if the neighbourhood given
        to the constructor is a MultiNeighbourhood.
        If this function is called when the first neighbourhood is the current
        neighbourhood, the first neighbourhood will remain the current
        neighbourhood.

        Raises
        ------
        WrongMoveTypeError
            If the move_function isn't a MultiNeighbourhood.

        """

        raise NotImplementedError

    def select_get_moves(self):
        """Function to get all moves from the current neighbourhood.

        Note that this function will only be useable if the neighbourhood given
        to the constructor is a MultiNeighbourhood.

        Returns
        -------
        generator
            An iterable generator object that contains all the moves of the
            current neighbourhood.

        Raises
        ------
        WrongMoveTypeError
            If the move_function isn't a MultiNeighbourhood.

        """

        raise NotImplementedError

    def select_random_move(self):
        """A method used to generate a random move from the current neighbourhood.

        Note that this function will only be useable if the neighbourhood given
        to the constructor is a MultiNeighbourhood.

        Returns
        -------
        tuple of int
            A random valid move from the current neighbourhood.

        Raises
        ------
        WrongMoveTypeError
            If the move_function isn't a MultiNeighbourhood.

        """

        raise NotImplementedError

    @abstractmethod
    def reset(self):
        """Resets the object back to it's state after init."""

        pass



from numpy import array


def convert_data(data):
    """Converts a list made with add_to_data_func into seperate lists.

    Parameters
    ----------
    data : list of tuple
        A list created with add_to_data_func

    Returns
    -------
    tuple of numpy.ndarray
        All the data in seperate lists. The first list will always be the time
        passed.

    """

    size = len(data[0])

    all_lists = []

    for i in range(size):
        all_lists.append([])

    # convert to lists
    for data_point in data:
        for i in range(size):
            all_lists[i].append(data_point[i])


    # convert timestamps into the time passed
    offset = all_lists[0][0]
    for i in range(len(all_lists[0])):
        all_lists[0][i] -= offset

    # create tuple and convert lists to ndarray

    all_data = ()

    for l in all_lists:
        all_data = all_data + (array(l),)

    return all_data



from lclpy.localsearch.abstract_local_search import AbstractLocalSearch
from lclpy.termination.always_true_criterion import AlwaysTrueCriterion

from lclpy.aidfunc.is_improvement_func import bigger, smaller
from lclpy.aidfunc.pass_func import pass_func
from lclpy.aidfunc.add_to_data_func import add_to_data_func
from lclpy.aidfunc.convert_data import convert_data
from lclpy.aidfunc.logging import log_improvement

from collections import namedtuple


class SteepestDescent(AbstractLocalSearch):
    """Performs a steepest descent algorithm on the given problem.

    Parameters
    ----------
    problem : AbstractLocalSearchProblem
        Contains all the data needed for the specific problem.
    minimise : bool, optional
        If the goal is to minimise the evaluation function, this should be
        True. If the goal is to maximise the evaluation function, this should
        be False. The default is True.
    termination_criterion : AbstractTerminationCriterion, optional
        The termination criterion that is used.
    benchmarking : bool, optional
        Should be True if one wishes benchmarks to be kept, should be False if
        one wishes no benchmarks to be made. Default is False.
    logging: bool, optional
        Improvements will be logged to the command line if this variable is
        True. Default is True.

    Attributes
    ----------
    _problem : AbstractLocalSearchProblem
        Contains all the data needed for the specific problem.
    _termination_criterion : AbstractTerminationCriterion
        Ends the algorithm if no more improvements can be found.
    _function
        The function used to determine if a delta value is better than another
        delta value.
    _best_found_delta_base_value : float
        Initialisation value for the delta value of each iteration. It's
        infinite when minimising or minus infinite when maximising.
    data : list of tuple
        Data useable for benchmarking will be None if no benchmarks are made.
    _data_append
        Function to append new data-points to data. Will do nothing if no
        benchmarks are made.
    _log_improvement
        Function to write logs to the command line. Will do nothing if no logs
        are made.

    Examples
    --------
    An example of minimising:

    .. doctest::

        >>> import numpy
        >>> from lclpy.localsearch.steepestdescent.steepest_descent \\
        ...     import SteepestDescent
        >>> from lclpy.localsearch.move.tsp_array_swap import TspArraySwap
        >>> from lclpy.evaluation.tsp_evaluation_function \\
        ...     import TspEvaluationFunction
        >>> from lclpy.problem.array_problem import ArrayProblem
        ... # init problem
        >>> distance_matrix = numpy.array(
        ... [[0, 2, 5, 8],
        ...  [2, 0, 4, 1],
        ...  [5, 4, 0, 7],
        ...  [8, 1, 7, 0]])
        >>> size = distance_matrix.shape[0]
        >>> move = TspArraySwap(size)
        >>> evaluation = TspEvaluationFunction(distance_matrix, move)
        >>> problem = ArrayProblem(evaluation, move, size)
        ... # init SteepestDescent
        >>> steepest_descent = SteepestDescent(problem, logging=False)
        ... # run algorithm
        >>> steepest_descent.run()
        Results(best_order=array([0, 1, 3, 2]), best_value=15, data=None)

    An example of maximising, note that the distance matrix is different:

    .. doctest::

        >>> import numpy
        >>> from lclpy.localsearch.steepestdescent.steepest_descent \\
        ...     import SteepestDescent
        >>> from lclpy.localsearch.move.tsp_array_swap import TspArraySwap
        >>> from lclpy.evaluation.tsp_evaluation_function \\
        ...     import TspEvaluationFunction
        >>> from lclpy.problem.array_problem import ArrayProblem
        ... # init problem
        >>> distance_matrix = numpy.array(
        ... [[0, 8, 5, 2],
        ...  [8, 0, 4, 7],
        ...  [5, 4, 0, 1],
        ...  [2, 7, 1, 0]])
        >>> size = distance_matrix.shape[0]
        >>> move = TspArraySwap(size)
        >>> evaluation = TspEvaluationFunction(distance_matrix, move)
        >>> problem = ArrayProblem(evaluation, move, size)
        ... # init SteepestDescent
        >>> steepest_descent = SteepestDescent(problem, False, logging=False)
        ... # run algorithm
        >>> steepest_descent.run()
        Results(best_order=array([0, 1, 3, 2]), best_value=21, data=None)



    """

    def __init__(self, problem, minimise=True, termination_criterion=None,
                 benchmarking=False, logging=True):
        super().__init__()

        self._problem = problem

        if termination_criterion is None:
            self._termination_criterion = AlwaysTrueCriterion()
        else:
            self._termination_criterion = termination_criterion

        if minimise:
            self._function = smaller
            self._best_found_delta_base_value = float("inf")
        else:
            self._function = bigger
            self._best_found_delta_base_value = float("-inf")

        if benchmarking:
            self.data = []
            self._data_append = add_to_data_func
        else:
            self.data = None
            self._data_append = pass_func

        if logging:
            self._log_improvement = log_improvement
        else:
            self._log_improvement = pass_func

    def run(self):
        """Starts running the steepest descent.

        Returns
        -------
        best_order : numpy.ndarray
            The best found order.
        best_value : int or float
            The evaluation value of the best found order.
        data : None or collections.namedtuple
            Data useable for benchmarking. If no benchmarks were made, it will
            be None. The namedtuple contains the following data in
            numpy.ndarrays:
            time, iteration, value


        """

        # init problem
        base_value = self._problem.evaluate()
        self._problem.set_as_best(base_value)

        # init iteration (used to vount the amount of iterations)
        iteration = 0

        # add to data
        self._data_append(self.data, iteration, base_value)

        # init termination criterion
        self._termination_criterion.check_first_value(base_value)
        self._termination_criterion.start_timing()

        while self._termination_criterion.keep_running():

            # search the neighbourhood for the best move
            best_found_delta = self._best_found_delta_base_value
            best_found_move = None

            for move in self._problem.get_moves():

                # check quality move
                delta = self._problem.evaluate_move(move)

                # keep data best move
                if self._function(best_found_delta, delta):
                    best_found_delta = delta
                    best_found_move = move

            # check if the best_found_move improves the delta, if this is the
            # case perform the move and set a new best problem
            base_value = base_value + best_found_delta

            self._termination_criterion.check_new_value(base_value)

            if self._function(self._problem.best_order_value, base_value):
                # if the found value is better

                # perform move and set new state as best.
                self._problem.move(best_found_move)
                self._problem.set_as_best(base_value)

                # log if needed
                self._log_improvement(base_value)

                # add to data
                self._data_append(self.data, iteration, base_value)

            else:

                # if it's worse: stop iterating by breaking from the loop
                break

            iteration += 1
            self._termination_criterion.iteration_done()

        # last data point
        self._data_append(self.data, iteration, base_value)

        # if we have data:
        # convert data to something easier to plot
        if self.data is not None:

            # convert to tuple of list
            data = convert_data(self.data)

            # make namedtuple
            DataAsLists = namedtuple('Data', ['time', 'iteration', 'value'])

            data = DataAsLists(data[0], data[1], data[2])

        else:
            data = None

        # return results

        Results = namedtuple('Results', ['best_order', 'best_value', 'data'])

        return Results(self._problem.best_order,
                       self._problem.best_order_value,
                       data)

    def reset(self):
        """Resets the object back to it's state after init.

        Raises
        ------
        NotImplementedError
            If the problem or termination criterion have no reset method
            implemented.

        Examples
        --------
        An example of minimising, reset after run:

        .. doctest::

            >>> import numpy
            >>> from lclpy.localsearch.steepestdescent.steepest_descent \\
            ...     import SteepestDescent
            >>> from lclpy.localsearch.move.tsp_array_swap \\
            ...     import TspArraySwap
            >>> from lclpy.evaluation.tsp_evaluation_function \\
            ...     import TspEvaluationFunction
            >>> from lclpy.problem.array_problem import ArrayProblem
            ... # init problem
            >>> distance_matrix = numpy.array(
            ... [[0, 2, 5, 8],
            ...  [2, 0, 4, 1],
            ...  [5, 4, 0, 7],
            ...  [8, 1, 7, 0]])
            >>> size = distance_matrix.shape[0]
            >>> move = TspArraySwap(size)
            >>> evaluation = TspEvaluationFunction(distance_matrix, move)
            >>> problem = ArrayProblem(evaluation, move, size)
            ... # init SteepestDescent
            >>> steepest_descent = SteepestDescent(problem, logging=False)
            ... # state before running
            >>> steepest_descent._problem._order
            array([0, 1, 2, 3])
            >>> steepest_descent._termination_criterion.keep_running()
            True
            >>> # run algorithm
            >>> steepest_descent.run()
            Results(best_order=array([0, 1, 3, 2]), best_value=15, data=None)
            >>> # tests reset
            >>> # before reset
            >>> steepest_descent._problem._order
            array([0, 1, 3, 2])
            >>> steepest_descent._termination_criterion.keep_running()
            True
            >>> # reset
            >>> steepest_descent.reset()
            ... # after reset
            >>> steepest_descent._problem._order
            array([0, 1, 2, 3])
            >>> steepest_descent._termination_criterion.keep_running()
            True


        """

        self._problem.reset()
        self._termination_criterion.reset()

        if self.data is not None:
            self.data = []



from abc import ABC, abstractmethod


class AbstrIterationsTempFunction(ABC):
    """Template class for determining the number of iterations for a temperature.

    Implementations of this class are meant to be used in a simulated annealing
    algorithm. Implementations are supposed to be used to determine the amount
    of iterations for a certain temperature.

    """

    def __init__(self):
        super().__init__()

    @abstractmethod
    def get_iterations(self, temperature):
        """Returns the amount of iterations for a certain temperature.

        Parameters
        ----------
        Temperature : int or float
            The current temperature.

        """

        pass



import matplotlib.pyplot as plt


def plot(data, title='Time - evaluation value plot.', width=20, height=10):
    """Plots the data in a time-value plot.

    Parameters
    ----------
    data : collections.namedtuple
        The data tuple one gets by performing a localsearch algorithm.
    title : str, optional
        The title of the plot. \n
        Default is 'Time - evaluation value plot.'.
    width : int, optional
        The width of the plot.
    height : int, optional
        The height of the plot.

    """

    # get the data from the data object
    time = data.time
    values = data.value

    plt.figure(figsize=(width, height))
    plt.plot(time, values, label='value')

    if hasattr(data, 'best_value'):
        best_values = data.best_value

        plt.plot(time, best_values, label='best value found')

    plt.xlabel('time (s)')
    plt.ylabel('evaluation value')

    plt.title(title)

    plt.legend()

    plt.show()


def iterations_plot(data, title='Iterations - evaluation value plot.',
                    width=20, height=10):
    """Plots the data in an iterations-value plot.

    Parameters
    ----------
    data : collections.namedtuple
        The data tuple one gets by performing a localsearch algorithm.
    title : str, optional
        The title of the plot. \n
        Default is 'Iterations - evaluation value plot.'.
    width : int, optional
        The width of the plot.
    height : int, optional
        The height of the plot.

    """

    # get the data from the data object
    iterations = data.iteration
    values = data.value

    plt.figure(figsize=(width, height))
    plt.plot(iterations, values, label='value')

    if hasattr(data, 'best_value'):
        best_values = data.best_value

        plt.plot(iterations, best_values, label='best value found')

    plt.xlabel('iterations')
    plt.ylabel('evaluation value')

    plt.title(title)

    plt.legend()

    plt.show()


def plot_single_stat(benchmark_single_stat, title='Single stat table',
                     algorithm_names=None, problem_names=None, width=10,
                     height=None):
    """Plots a single statistic of a benchmark.

    Parameters
    ----------
    benchmark_single_stat : numpy.ndarray
        A single 2D array that is the result from a function in
        "lclpy.benchmark.statistics".
    algorithm_names : list of str or tuple of str, optional
        Must be the same length as the algorithm list given to the original
        benchmark. The indices of a name corresponds with the indices of the
        algorithms given to the original benchmark.
    problem_names : list of str or tuple of str, optional
        Must be the same length as the problem list given to the original
        benchmark. The indices of a name corresponds with the indices of the
        problems given to the original benchmark.
    width : float, optional
        The width of the table in inches. Default is 10.
    height : float, optional
        The height of the table in inches. The default scales with the amount
        of items in the table.

    """

    if algorithm_names is None:
        algorithm_names = tuple(range(len(benchmark_single_stat)))

    if problem_names is None:
        problem_names = tuple(range(len(benchmark_single_stat[0])))

    fig, ax = plt.subplots()

    ax.axis('off')

    fig.set_figwidth(width)

    if height is None:
        vert_size = len(benchmark_single_stat)
        fig.set_figheight(vert_size * 0.2 + 1.2)
    else:
        fig.set_figheight(height)

    table = ax.table(cellText=benchmark_single_stat, rowLabels=algorithm_names,
                     colLabels=problem_names, loc='center')

    table.set_fontsize(14)
    table.scale(1.5, 1.5)

    plt.title(title, fontsize=20)
    plt.show()



from lclpy.termination.abstract_termination_criterion \
    import AbstractTerminationCriterion


class AlwaysTrueCriterion(AbstractTerminationCriterion):
    """Termination criterion that will never terminate an algorithm.

    This criterion is only meant to be used as a default termination criterion
    in algorithms that will stop iterating when no improvement is found. Don't
    use it in any other case.

    """

    def __init__(self):
        super().__init__()

    def keep_running(self):
        """function to determine if the algorithm needs to continue running.

        Returns
        -------
        bool
            Will always be True.

        """

        return True

    def reset(self):
        """Resets the object back to it's state after init."""

        pass
